# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import unittest

from enum import Enum

from hamcrest import *

from samitorch.factories.factories import MetricsFactory
from samitorch.factories.enums import Metrics


class IncorrectMetrics(Enum):
    Incorrect = "Incorrect"


class ModelFactoryTest(unittest.TestCase):

    def setUp(self):
        self.metric_factory = MetricsFactory()

    def test_should_instantiate_dice_metric(self):
        metric = self.metric_factory.create_metric(Metrics.Dice, num_classes=3)
        assert_that(metric, is_not(None))



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import abc


class Gauge(object):
    """Base class for all Metrics."""

    def __init__(self):
        """Class constructor."""
        self.reset()

    @abc.abstractmethod
    def update(self, value: float, **kwargs):
        """Update the gauge's state.

        Args:
            value (float): The value with which to update the gauge.
            **kwargs (dict): keyword arguments.

        Raises:
            NotImplementedError: if not overwritten by subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def reset(self):
        """Reset gauge's data.

        Raises:
            NotImplementedError: if not overwritten by subclass.
        """
        raise NotImplementedError()


class RunningAverageGauge(Gauge):
    """A moving accuracy gauge."""

    def __init__(self) -> None:
        """Class constructor."""
        self.count = 0
        self.sum = 0
        self.average = 0
        super(RunningAverageGauge).__init__()

    def update(self, value: float, n_data: int = 1) -> None:
        """Update the moving average.

        Args:
            value (float): The value with which to update the average.
            n_data (int): The number of data passed to this moving average.
        """
        self.count += n_data
        self.sum += value * n_data
        self.average = self.sum / self.count

    def reset(self) -> None:
        """Reset the moving average."""
        self.count = 0
        self.sum = 0
        self.average = 0



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import torch
import numpy as np

from typing import Union, Tuple
from torch.utils.data._utils.collate import default_collate
from torch.utils.data.sampler import SubsetRandomSampler


class DataLoader(torch.utils.data.DataLoader):
    """
    Base class for all data loaders.
    """

    def __init__(self, dataset: torch.utils.data.dataset.Dataset, batch_size: int, shuffle: bool,
                 validation_split: Union[int, float], num_workers: int, collate_fn=default_collate,
                 samplers: Union[tuple, list] = None):
        """
        DataLoader initializer.

        Args:
            dataset (:obj:`torch.utils.data.dataset.Dataset`): A PyTorch Dataset object which contains the training data.
            batch_size (int): The number of elements to load at each batch.
            shuffle (bool): Whether to shuffle the batch or not.
            validation_split (int_or_float): If int, takes this number of elements to produce a validation set.
                If float, computes the proportion of the training set to place in a validation set.
            num_workers (int): Number of parallel workers to execute.
            collate_fn (callable): The collate function that merges a list of samples to form a mini-batch.
            samplers: Optional list or tuple containing both a training and validation sampler.
        """
        self._validation_split = validation_split
        self._shuffle = shuffle
        self._batch_idx = 0
        self._n_samples = len(dataset)

        if samplers is None:
            self._sampler, self._valid_sampler = self._split_sampler(self._validation_split)
        else:
            self._sampler, self._valid_sampler = samplers
            self._shuffle = False

        self._init_kwargs = {
            'dataset': dataset,
            'batch_size': batch_size,
            'shuffle': self._shuffle,
            'collate_fn': collate_fn,
            'num_workers': num_workers,
            'pin_memory': torch.cuda.is_available()
        }
        super().__init__(sampler=self._sampler, **self._init_kwargs)

    def _split_sampler(self, split: Union[int, float]) -> Tuple[Union[
                                                                    torch.utils.data.sampler.Sampler, None], Union[
                                                                    torch.utils.data.sampler.Sampler, None]]:
        """
        Split a sampler for training and validation split.

        Args:
            split (int_or_float): If int, takes this number of elements to produce a validation set.
                If float, computes the proportion of the training set to place in a validation set.

        Returns:
            Tuple_of_:obj:`torch.utils.data.sampler.Sampler`: A tuple of samplers for training and validation sets.
        """
        if split == 0.0:
            return None, None

        idx_full = np.arange(self._n_samples)

        np.random.shuffle(idx_full)

        if isinstance(split, int):
            assert split > 0
            assert split < self._n_samples, "Validation set size is configured to be larger than entire dataset."
            len_valid = split
        else:
            len_valid = int(self._n_samples * split)

        valid_idx = idx_full[0:len_valid]
        train_idx = np.delete(idx_full, np.arange(0, len_valid))

        train_sampler = SubsetRandomSampler(train_idx)
        valid_sampler = SubsetRandomSampler(valid_idx)

        # Turn off shuffle option which is mutually exclusive with sampler
        self._shuffle = False
        self._n_samples = len(train_idx)

        return train_sampler, valid_sampler

    def get_validation_dataloader(self) -> Union[torch.utils.data.dataloader.DataLoader, None]:
        """
        Get the validation dataloader object based on previously created validation sampler.

        Returns:
            :obj:`torch.utils.data.dataloader.Dataloader`: A PyTorch DataLoader with the validation sampler (if exists),
                else returns None.
        """
        if self._valid_sampler is None:
            return None
        else:
            return torch.utils.data.dataloader.DataLoader(sampler=self._valid_sampler, **self._init_kwargs)



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from enum import Enum


class ActivationLayers(Enum):
    ReLU = "ReLU"
    LeakyReLU = "LeakyReLU"
    PReLU = "PReLU"


class PaddingLayers(Enum):
    ReplicationPad3d = "ReplicationPad3d"


class PoolingLayers(Enum):
    MaxPool3d = "MaxPool3d"
    AvgPool3d = "AvgPool3d"
    Conv3d = "Conv3d"


class NormalizationLayers(Enum):
    GroupNorm = "GroupNorm"
    BatchNorm3d = "BatchNorm3d"


class ResNetModels(Enum):
    ResNet18 = "ResNet18"
    ResNet34 = "ResNet34"
    ResNet50 = "ResNet50"
    ResNet101 = "ResNet101"
    ResNet152 = "ResNet152"


class UNetModels(Enum):
    UNet3D = "UNet3D"


class Optimizers(Enum):
    SGD = "SGD"
    Adam = "Adam"
    Adagrad = "Adagrad"
    Adadelta = "Adadelta"
    SparseAdam = "SparseAdam"
    Adamax = "Adamax"
    Rprop = "Rprop"
    RMSprop = "RMSprop"
    ASGD = "ASGD"


class Metrics(Enum):
    Dice = "Dice"
    GeneralizedDice = "GeneralizeDice"
    Accuracy = "Accuracy"
    Precision = "Precision"
    MeanAbsoluteError = "MeanAbsoluteError"
    MeanPairwiseDistance = "MeanPairwiseDistance"
    MeanSquaredError = "MeanSquaredError"
    Recall = "Recall"
    RootMeanSquaredError = "RootMeanSquaredError"
    TopKCategoricalAccuracy = "TopKCategoricalAccuracy"
    IoU = "IoU"
    mIoU = "mIoU"



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import abc
import yaml
import logging

from samitorch.configs.configurations import UNetModelConfiguration, ResNetModelConfiguration, ModelConfiguration


class AbstractConfigurationParserFactory(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def parse(self, path: str):
        raise NotImplementedError

    @abc.abstractmethod
    def register(self, model_type: str, configuration_class):
        raise NotImplementedError


class ModelConfigurationParserFactory(AbstractConfigurationParserFactory):

    def __init__(self) -> None:
        self._supported_model_configuration = {
            "UNet3D": UNetModelConfiguration,
            "ResNet3D": ResNetModelConfiguration
        }

    def parse(self, path: str) -> ModelConfiguration:
        """
        Parse a model configuration file.

        Args:
            path (str): Configuration YAML file path.

        Returns:
            :obj:`ModelConfiguration`: An object containing model's properties.

        """
        with open(path, 'r') as config_file:
            try:
                config = yaml.load(config_file, Loader=yaml.FullLoader)
                model_type = config["model"]["type"]
                model_configuration = self._supported_model_configuration.get(model_type)
                return model_configuration(config["model"])
            except ValueError as e:
                logging.error("Model type {} not supported.".format(model_type, e))
            except yaml.YAMLError as e:
                logging.error(
                    "Unable to read the training config file: {} with error {}".format(path, e))

    def register(self, model_type: str, model_configuration_class: ModelConfiguration) -> None:
        """
        Register a new type of model.

        Args:
            model_type (str): The generic model type (e.g. 'unet3d' or 'resnet3d').
            model_configuration_class (:obj:`samitorch.config.model.ModelConfiguration`): The class defining model's
                properties.
        """
        self._supported_model_configuration[model_type] = model_configuration_class



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import torch
import unittest
import numpy as np

from samitorch.factories.parsers import ModelConfigurationParserFactory
from samitorch.models.unet3d import UNet3D, SingleConv, DoubleConv, Encoder, Decoder
from tests.models.model_helper_test import TestModelHelper
from samitorch.factories.enums import UNetModels, ActivationLayers

from samitorch.factories.factories import ModelFactory


class Unet3DTest(unittest.TestCase):
    CONFIGURATION_PATH = "samitorch/configs/unet3d.yaml"

    def setUp(self):
        self.configurationParserFactory = ModelConfigurationParserFactory()
        self.config = self.configurationParserFactory.parse(self.CONFIGURATION_PATH)
        self.factory = ModelFactory()
        self.input = torch.rand((2, 1, 32, 32, 32))
        self.y = torch.randint(low=0, high=2, size=(2, 1, 32, 32, 32)).float()

    def test_model_should_be_created_with_config(self):
        model = UNet3D(self.config)
        assert isinstance(model, UNet3D)

    def test_model_should_update_vars(self):
        model = self.factory.create_model(UNetModels.UNet3D, self.config)
        helper = TestModelHelper(model, torch.nn.BCEWithLogitsLoss(),
                                 torch.optim.SGD(model.parameters(), lr=0.01))
        helper.assert_vars_change((self.input, self.y))
        output = model.forward(self.input)
        assert isinstance(output, torch.Tensor)

    def test_model_output_should_have_same_dimensions_than_input(self):
        input_dim = np.array(list(self.input.size()))
        model = self.factory.create_model(UNetModels.UNet3D, self.config)
        output = model.forward(self.input)
        output_dim = np.array(list(output.size()))
        np.testing.assert_array_equal(input_dim, output_dim)


class UNet3DModulesTest(unittest.TestCase):
    CONFIGURATION_PATH = "samitorch/configs/unet3d.yaml"

    def setUp(self):
        self.configurationParserFactory = ModelConfigurationParserFactory()
        self.config = self.configurationParserFactory.parse(self.CONFIGURATION_PATH)
        self.input = torch.rand((2, 1, 32, 32, 32))
        self.input_high_channels = torch.rand(2, 128, 16, 16, 16)
        self.in_channels = 1
        self.out_channels = 64
        self.kernel_size = 3
        self.num_groups = 8
        self.padding = (1, 1, 1, 1, 1, 1)
        self.relu_activation = ActivationLayers.ReLU
        self.leaky_relu_activation = ActivationLayers.LeakyReLU

    def test_single_conv_with_padding_should_give_correct_dimensions(self):
        block = SingleConv(self.in_channels, self.out_channels, self.kernel_size, self.num_groups, self.padding,
                           self.relu_activation)
        output = block(self.input)
        assert output.size() == torch.Size([2, 64, 32, 32, 32])

    def test_single_conv_without_padding_should_give_correct_dimensions(self):
        block = SingleConv(self.in_channels, self.out_channels, self.kernel_size, self.num_groups, None,
                           self.relu_activation)
        output = block(self.input)
        assert output.size() == torch.Size([2, 64, 32, 32, 32])

    def test_single_conv_with_leaky_relu_should_give_correct_instance(self):
        block = SingleConv(self.in_channels, self.out_channels, self.kernel_size, self.num_groups, None,
                           self.leaky_relu_activation)
        assert isinstance(block._activation, torch.nn.LeakyReLU)

    def test_single_conv_with_relu_should_give_correct_instance(self):
        block = SingleConv(self.in_channels, self.out_channels, self.kernel_size, self.num_groups, None,
                           self.relu_activation)
        assert isinstance(block._activation, torch.nn.ReLU)

    def test_single_conv_without_activation_should_instantiate_correctly(self):
        block = SingleConv(self.in_channels, self.out_channels, self.kernel_size, self.num_groups, self.padding,
                           None)
        assert block is not None
        assert block._activation is None

    def test_single_conv_without_group_norm_should_give_correct_instance(self):
        block = SingleConv(self.in_channels, self.out_channels, self.kernel_size, None, None,
                           self.relu_activation)
        assert isinstance(block._norm, torch.nn.BatchNorm3d)

    def test_double_conv_should_give_correct_dimensions(self):
        block = DoubleConv(self.in_channels, self.out_channels, True, self.kernel_size, self.num_groups, self.padding,
                           self.relu_activation)
        output = block(self.input)
        assert output.size() == torch.Size([2, 64, 32, 32, 32])

    def test_encoder_should_give_correct_pooling(self):
        encoder = Encoder(self.in_channels, self.out_channels, DoubleConv, self.config)
        assert hasattr(encoder, "_pooling")
        assert isinstance(encoder._pooling, torch.nn.MaxPool3d)

    def test_encoder_should_give_correct_dimension(self):
        encoder = Encoder(self.in_channels, self.out_channels, DoubleConv, self.config)
        output = encoder(self.input)
        assert output.size() == torch.Size([2, 64, 16, 16, 16])

    def test_decoder_should_use_upsampling(self):
        decoder = Decoder(self.out_channels, self.in_channels, DoubleConv, self.config)
        assert hasattr(decoder, "_upsample")
        assert decoder._upsample is None



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import abc


class AbstractPreProcessingPipeline(metaclass=abc.ABCMeta):
    """
    Define a preprocessing pipeline.
    """

    @abc.abstractmethod
    def run(self, **kwargs):
        """
        Run the preprocessing pipeline.
        Args:
            **kwargs: Optional keyword arguments.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import unittest
import torch
import os

from hamcrest import *

from samitorch.utils.utils import to_onehot, flatten, extract_file_paths


class UtilsTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")

    def setUp(self):
        pass

    @staticmethod
    def test_to_onehot():
        indices = torch.LongTensor([0, 1, 2, 3])
        actual = to_onehot(indices, 4)
        expected = torch.eye(4)
        assert actual.equal(expected)

        y = torch.randint(0, 21, size=(1000,))
        y_ohe = to_onehot(y, num_classes=21)
        y2 = torch.argmax(y_ohe, dim=1)
        assert y.equal(y2)

        y = torch.randint(0, 21, size=(4, 250, 255))
        y_ohe = to_onehot(y, num_classes=21)
        y2 = torch.argmax(y_ohe, dim=1)
        assert y.equal(y2)

        y = torch.randint(0, 21, size=(4, 150, 155, 4, 6))
        y_ohe = to_onehot(y, num_classes=21)
        y2 = torch.argmax(y_ohe, dim=1)
        assert y.equal(y2)

    @staticmethod
    def test_flatten():
        x = torch.randint(0, 255, size=(15, 3, 10, 10))
        flattened_x = flatten(x)
        assert flattened_x.size() == torch.Size([3, 1500])

    def test_glob_imgs(self):
        all_paths = [os.path.join(self.TEST_DATA_FOLDER_PATH, "DTI.nii"),
                     os.path.join(self.TEST_DATA_FOLDER_PATH, "FA.nii"),
                     os.path.join(self.TEST_DATA_FOLDER_PATH, "Mask.nii"),
                     os.path.join(self.TEST_DATA_FOLDER_PATH, "T1.nii"),
                     os.path.join(self.TEST_DATA_FOLDER_PATH, "T1_1mm.nii")]

        paths = extract_file_paths(self.TEST_DATA_FOLDER_PATH)

        assert_that(paths, equal_to(all_paths))



#  -*- coding: utf-8 -*-
#  Copyright 2019 SAMITorch Authors. All Rights Reserved.
#  #
#  Licensed under the MIT License;
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://opensource.org/licenses/MIT
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#  ==============================================================================

import abc
import torch

from enum import Enum
from typing import Union

from samitorch.configs.configurations import ModelConfiguration
from samitorch.factories.enums import *


class AbstractLayerFactory(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def create_layer(self, *args):
        raise NotImplementedError

    @abc.abstractmethod
    def register(self, function: str, creator: torch.nn.Module):
        raise NotImplementedError


class AbstractCriterionFactory(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def create_criterion(self, criterion: Union[str, Enum], *args):
        raise NotImplementedError

    @abc.abstractmethod
    def register(self, criterion: str, creator):
        raise NotImplementedError


class AbstractMetricFactory(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def create_metric(self, function: Union[str, Metrics], *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def register(self, function: str, creator):
        raise NotImplementedError


class AbstractModelFactory(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def create_model(self, name: Union[str, ResNetModels, UNetModels], config: ModelConfiguration):
        pass

    @abc.abstractmethod
    def register(self, model: str, creator):
        raise NotImplementedError


class AbstractOptimizerFactory(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def create_optimizer(self, function: Union[str, Optimizers], *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def register(self, function: str, creator):
        raise NotImplementedError


class OptimizerFactory(AbstractOptimizerFactory):

    def __init__(self):
        super(OptimizerFactory, self).__init__()

        self._optimizers = {
            "Adam": torch.optim.Adam,
            "Adagrad": torch.optim.Adagrad,
            "SGD": torch.optim.SGD,
            "Adadelta": torch.optim.Adadelta,
            "SparseAdam": torch.optim.SparseAdam,
            "Adamax": torch.optim.Adamax,
            "Rprop": torch.optim.Rprop,
            "RMSprop": torch.optim.RMSprop,
            "ASGD": torch.optim.ASGD
        }

    def create_optimizer(self, function: Union[str, Optimizers], *args, **kwargs):
        """
        Instanciate an optimizer based on its name.

        Args:
            function (Enum): The optimizer's name.
            *args: Other arguments.

        Returns:
            :obj:`torch.optim.Optimizer`: The optimizer.

        Raises:
            KeyError: Raises KeyError Exception if Activation Function is not found.
        """
        optimizer = self._optimizers[function.name if isinstance(function, Optimizers) else function]
        return optimizer(*args, **kwargs)

    def register(self, function: str, creator: torch.optim.Optimizer):
        """
        Add a new activation layer.

        Args:
           function (str): Activation layer name.
           creator (:obj:`torch.nn.Module`): A torch module object wrapping the new custom optimizer function.
        """
        self._optimizers[function] = creator


class ModelFactory(AbstractModelFactory):
    """
    Object to instantiate a model.
    """

    def __init__(self):
        from samitorch.models.resnet3d import ResNet18, ResNet34, ResNet50, ResNet101, ResNet152
        from samitorch.models.unet3d import UNet3D
        self._models = {
            'ResNet18': ResNet18,
            'ResNet34': ResNet34,
            'ResNet50': ResNet50,
            'ResNet101': ResNet101,
            'ResNet152': ResNet152,
            'UNet3D': UNet3D
        }

    def create_model(self, model_name: Union[str, Enum], config: ModelConfiguration) -> torch.nn.Module:
        """
        Instantiate a new support model.

        Args:
            model_name (str): The model's name (e.g. 'UNet3D').
            config (:obj:`samitorch.configs.model_configuration.ModelConfiguration`): An object containing model's parameters.

        Returns:
            :obj:`torch.nn.Module`: A PyTorch model.
        """
        model = self._models.get(model_name.name if isinstance(model_name, Enum) else model_name)
        if not model:
            raise ValueError("Model {} is not supported.".format(model_name))
        return model(config)

    def register(self, model: str, creator):
        """
        Add a new model.

        Args:
          model (str): Model's name.
          creator: A torch module object wrapping the new custom model.
        """
        self._models[model] = creator


class MetricsFactory(AbstractMetricFactory):
    def __init__(self):
        import ignite
        from samitorch.metrics.metrics import Dice

        super(MetricsFactory, self).__init__()
        self._metrics = {
            "Dice": Dice,
            "Accuracy": ignite.metrics.Accuracy,
            "Precision": ignite.metrics.Precision,
            "MeanAbsoluteError": ignite.metrics.MeanAbsoluteError,
            "MeanPairwiseDistance": ignite.metrics.MeanPairwiseDistance,
            "MeanSquaredError": ignite.metrics.MeanSquaredError,
            "Recall": ignite.metrics.Recall,
            "RootMeanSquaredError": ignite.metrics.RootMeanSquaredError,
            "TopKCategoricalAccuracy": ignite.metrics.TopKCategoricalAccuracy,
            "IoU": ignite.metrics.IoU,
            "mIoU": ignite.metrics.mIoU
        }

    def create_metric(self, metric: Union[str, Metrics], *args, **kwargs):
        """
        Instanciate an optimizer based on its name.

        Args:
            metric (str_or_Enum): The optimizer's name.
            *args: Other arguments.

        Returns:
            The metric, a torch or ignite module.

        Raises:
            KeyError: Raises KeyError Exception if Activation Function is not found.
        """
        metric = self._metrics[metric.name if isinstance(metric, Metrics) else metric]
        return metric(*args, **kwargs)

    def register(self, metric: str, creator):
        """
        Add a new activation layer.

        Args:
           metric (str): Metric's name.
           creator: A torch or ignite module object wrapping the new custom metric function.
        """
        self._metrics[metric] = creator


class ActivationLayerFactory(AbstractLayerFactory):
    """
    Object to instantiate an activation layer.
    """

    def __init__(self):
        super(ActivationLayerFactory, self).__init__()

        self._activation_functions = {
            "ReLU": torch.nn.ReLU,
            "LeakyReLU": torch.nn.LeakyReLU,
            'PReLU': torch.nn.PReLU
        }

    def create_layer(self, function: Union[str, ActivationLayers], *args, **kwargs):
        """
        Instantiate an activation layer based on its name.

        Args:
            function (str): The activation layer's name.
            *args: Optional arguments for the respective activation function.

        Returns:
            :obj:`torch.nn.Module`: The activation layer.

        Raises:
            KeyError: Raises KeyError Exception if Activation Function is not found.
        """
        activation_function = self._activation_functions[
            function.name if isinstance(function, ActivationLayers) else function]
        return activation_function(*args, **kwargs)

    def register(self, function: str, creator: torch.nn.Module):
        """
        Add a new activation layer.

        Args:
            function (str): Activation layer name.
            creator (:obj:`torch.nn.Module`): A torch module object wrapping the new custom activation function.
        """
        self._activation_functions[function] = creator


class PaddingLayerFactory(AbstractLayerFactory):
    """
    Object to instantiate a padding layer.
    """

    def __init__(self):
        super(PaddingLayerFactory, self).__init__()
        self._padding_strategies = {
            "ReplicationPad3d": torch.nn.ReplicationPad3d
        }

    def create_layer(self, strategy: Union[str, PaddingLayers], dims: tuple, *args, **kwargs):
        """
        Instantiate a new padding layer.

        Args:
            strategy (:obj:samitorch.models.layers.PaddingLayers): The padding strategy.
            dims (tuple): The number of  where to apply padding.
            *args: Optional arguments for the respective activation function.

        Returns:
            :obj:`torch.nn.Module`: The padding layer.

        Raises:
            KeyError: Raises KeyError Exception if Padding Function is not found.
        """
        padding = self._padding_strategies[strategy.name if isinstance(strategy, PaddingLayers) else strategy]
        return padding(dims, *args, **kwargs)

    def register(self, strategy: str, creator: torch.nn.Module):
        """
        Add a new padding strategy.

        Args:
            strategy (str): The padding strategy name.
            creator (:obj:`torch.nn.Module`): A torch module object wrapping the new custom padding layer.
        """
        self._padding_strategies[strategy] = creator


class PoolingLayerFactory(AbstractLayerFactory):
    """
    An object to instantiate pooling layers.
    """

    def __init__(self):
        super(PoolingLayerFactory, self).__init__()
        self._pooling_strategies = {
            "MaxPool3d": torch.nn.MaxPool3d,
            "AvgPool3d": torch.nn.AvgPool3d,
            "Conv3d": torch.nn.Conv3d
        }

    def create_layer(self, strategy: Union[str, PoolingLayers], kernel_size: int, stride: int = None, *args,
                     **kwargs):
        """
        Instantiate a new pooling layer with mandatory parameters.

        Args:
            strategy (str): The pooling strategy name.
            kernel_size (int): Pooling kernel's size.
            stride (int): The size of the stride of the window.
            *kwargs: Optional keywords.

        Returns:
            :obj:`torch.nn.Module`: The pooling layer.

        Raises:
            KeyError: Raises KeyError Exception if Pooling Function is not found.
        """
        pooling = self._pooling_strategies[strategy.name if isinstance(strategy, PoolingLayers) else strategy]

        if not "Conv3d" in strategy.name:
            return pooling(kernel_size, stride, *args, **kwargs)

        else:
            return pooling(*args, kernel_size, stride, **kwargs)

    def register(self, strategy: str, creator: torch.nn.Module):
        """
        Add a new pooling layer.

        Args:
            strategy (str): The pooling strategy name.
            creator (obj:`torch.nn.Module`): A torch module object wrapping the new custom pooling layer.
        """
        self._pooling_strategies[strategy] = creator


class NormalizationLayerFactory(AbstractLayerFactory):
    """
    An object to instantiate normalization layers.
    """

    def __init__(self):
        super(NormalizationLayerFactory, self).__init__()
        self._normalization_strategies = {
            "GroupNorm": torch.nn.GroupNorm,
            "BatchNorm3d": torch.nn.BatchNorm3d
        }

    def create_layer(self, strategy: Union[str, NormalizationLayers], *args, **kwargs):
        """
        Instantiate a new normalization layer.

        Args:
            strategy (str_or_Enum): The normalization strategy layer.
            *args: Optional keywords arguments.

        Returns:
            :obj:`torch.nn.Module`: The normalization layer.

        Raises:
            KeyError: Raises KeyError Exception if Normalization Function is not found.
        """
        norm = self._normalization_strategies[
            strategy.name if isinstance(strategy, NormalizationLayers) else strategy]
        return norm(*args, **kwargs)

    def register(self, strategy: str, creator: torch.nn.Module):
        """
        Add a new normalization layer.

        Args:
            strategy (str): The normalization strategy name.
            creator (:obj:`torch.nn.Module`): A torch module object wrapping the new custom normalization layer.
        """
        self._normalization_strategies[strategy] = creator


class CriterionFactory(AbstractCriterionFactory):
    def __init__(self):
        super(CriterionFactory, self).__init__()
        from samitorch.losses.losses import DiceLoss
        self._criterion = {
            "DiceLoss": DiceLoss,
            "CrossEntropyLoss": torch.nn.CrossEntropyLoss,
            "NLLLoss": torch.nn.NLLLoss,
            "BCEWithLogitsLoss": torch.nn.BCEWithLogitsLoss,
            "BCELoss": torch.nn.BCELoss,
            "MSELoss": torch.nn.MSELoss,
        }

    def create_criterion(self, criterion: Union[str, Enum], *args, **kwargs):
        """
        Instanciate a loss function based on its name.

        Args:
           criterion (str_or_Enum): The criterion's name.
           *args: Other arguments.

        Returns:
           :obj:`torch.nn.Module`: The criterion.

        Raises:
           KeyError: Raises KeyError Exception if Activation Function is not found.
        """
        criterion = self._criterion[criterion.name if isinstance(criterion, Enum) else criterion]
        return criterion(*args, **kwargs)

    def register(self, function: str, creator):
        """
        Add a new criterion.

        Args:
           function (str): Criterion's name.
           creator (:obj:`torch.nn.Module`): A torch module object wrapping the new custom criterion function.
        """
        self._criterion[function] = creator



# -*- coding: utf-8 -*-
# Coptargetsright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# targetsou matargets not use this file except in compliance with the License.
# You matargets obtain a coptargets of the License at
#
#     https://opensource.org/licenses/MIT
# Unless required btargets applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Define a metric computed during the training phase."""

import torch

from typing import Tuple

from ignite.metrics.confusion_matrix import ConfusionMatrix
from ignite.metrics import MetricsLambda, Metric

EPSILON = 1e-15


class Dice(Metric):
    """
    The Dice Metric.
    """

    def __init__(self, num_classes: int, reduction: str = None, average: str = None,
                 ignore_index: int = None, output_transform: callable = lambda x: x) -> None:
        """
        Metric initializer.

        Args:
            num_classes (int): The number of classes in the problem. In case of images, num_classes should also count the background index 0.
            average (str, optional): Confusion matrix values averaging schema: None, "samples", "recall", "precision".
                Default is None. If `average="samples"` then confusion matrix values are normalized by the number of seen
                samples. If `average="recall"` then confusion matrix values are normalized such that diagonal values
                represent class recalls. If `average="precision"` then confusion matrix values are normalized such that
                diagonal values represent class precisions.
            reduction (str): The type of reduction to apply (e.g. 'mean').
            ignore_index (int, optional): To ignore an index in Dice computation.
            output_transform (callable, optional): a callable that is used to transform the
                :class:`~ignite.engine.Engine`'s `process_function`'s output into the
                form expected by the metric. This can be useful if, for example, you have a multi-output model and
                you want to compute the metric with respect to one of the outputs.
        """
        self._num_classes = num_classes
        self._average = average
        self._ignore_index = ignore_index
        self._reduction = reduction
        self._cm = ConfusionMatrix(num_classes=num_classes, average=average,
                                   output_transform=output_transform)
        self._metric = compute_dice_coefficient(self._cm, self._ignore_index)

        if self._reduction == "mean":
            self._metric = self._metric.mean()

        super(Dice, self).__init__(output_transform=output_transform)

    def reset(self) -> None:
        """
        Reset the confusion matrix object.
        """
        self._cm.confusion_matrix = torch.zeros(self._num_classes, self._num_classes, dtype=torch.float64)

    def compute(self) -> torch.Tensor:
        """
        Compute the metric.

        Returns:
            :obj:`torch.Tensor`: The dice coefficient for each class.
        """
        return self._metric.compute()

    def update(self, output: Tuple[torch.Tensor, torch.Tensor]) -> None:
        """
        Update the confusion matrix with output values.

        Args:
            output (tuple of :obj:`torch.Tensor`): A tuple containing predictions and ground truth.
        """
        self._cm.update(output)


class GeneralizedDice(Metric):
    """
    The Generalized Dice Metric.
    """

    def __init__(self, num_classes: int, reduction: str = None, average: str = None,
                 ignore_index: int = None, output_transform: callable = lambda x: x) -> None:
        """
        Metric initializer.

        Args:
            num_classes (int): The number of classes in the problem. In case of images, num_classes should also count the background index 0.
            average (str, optional): Confusion matrix values averaging schema: None, "samples", "recall", "precision".
                Default is None. If `average="samples"` then confusion matrix values are normalized by the number of seen
                samples. If `average="recall"` then confusion matrix values are normalized such that diagonal values
                represent class recalls. If `average="precision"` then confusion matrix values are normalized such that
                diagonal values represent class precisions.
            reduction (str): The type of reduction to apply (e.g. 'mean').
            ignore_index (int, optional): To ignore an index in Dice computation.
            output_transform (callable, optional): a callable that is used to transform the
                :class:`~ignite.engine.Engine`'s `process_function`'s output into the
                form expected by the metric. This can be useful if, for example, you have a multi-output model and
                you want to compute the metric with respect to one of the outputs.
        """
        self._num_classes = num_classes
        self._average = average
        self._ignore_index = ignore_index
        self._metric = None
        self._reduction = reduction
        self._cm = ConfusionMatrix(num_classes=num_classes, average=average,
                                   output_transform=output_transform)
        super(GeneralizedDice, self).__init__(output_transform=output_transform)

    def reset(self) -> None:
        """
        Reset the confusion matrix object.
        """
        self._cm.confusion_matrix = torch.zeros(self._num_classes, self._num_classes, dtype=torch.float64)

    def compute(self) -> torch.Tensor:
        """
        Compute the metric.

        Returns:
            :obj:`torch.Tensor`: The dice coefficient for each class.
        """
        return self._metric.compute()

    def update(self, output: Tuple[torch.Tensor, torch.Tensor], weights: torch.Tensor = None) -> None:
        """
        Update the confusion matrix with output values.

        Args:
            output (tuple of :obj:`torch.Tensor`): A tuple containing predictions and ground truth.
            weights (:obj:`torch.Tensor`, optional): A weight vector which length equals to the number of classes.
        """

        self._metric = compute_generalized_dice_coefficient(self._cm, weights, self._ignore_index)

        if self._reduction == "mean":
            self._metric = self._metric.mean()

        self._cm.update(output)


def compute_dice_coefficient(cm: ConfusionMatrix, ignore_index: int = None):
    """
    Computes the Sørensen–Dice Coefficient (https://en.wikipedia.org/wiki/Sørensen–Dice_coefficient)

    Args:
        cm (:obj:`ignite.metrics.ConfusionMatrix`): A confusion matrix representing the classification of data.
        ignore_index (int): An index of a class to ignore for computation.

    Returns:
        array: The Sørensen–Dice Coefficient for each class.
    """
    if ignore_index is not None:
        validate_ignore_index(ignore_index)
        validate_num_classes(ignore_index, cm.num_classes)

    # Increase floating point precision
    cm = cm.type(torch.float64)
    dice = 2 * cm.diag() / (cm.sum(dim=1) + cm.sum(dim=0) + EPSILON)

    if ignore_index is not None:
        def remove_index(dice_vector):
            assert ignore_index <= len(dice_vector), "ignore_index {} is larger than the length " \
                                                     "of dice vector {}".format(ignore_index, len(dice_vector))
            indices = [x for x in range(len(dice_vector)) if x != ignore_index]
            return dice_vector[indices]

        return MetricsLambda(remove_index, dice)
    else:
        return dice


def compute_generalized_dice_coefficient(cm: ConfusionMatrix, weights: torch.Tensor, ignore_index: int = None):
    """
    Computes the Sørensen–Dice Coefficient (https://en.wikipedia.org/wiki/Sørensen–Dice_coefficient)

    Args:
        cm (:obj:`ignite.metrics.ConfusionMatrix`): A confusion matrix representing the classification of data.
        ignore_index (int): An index of a class to ignore for computation.
        weights (:obj:`torch.Tensor`): A weight vector which length equals to the number of classes.

    Returns:
        array: The Generalized Dice Coefficient for each class.
    """
    if ignore_index is not None:
        validate_ignore_index(ignore_index)
        validate_num_classes(ignore_index, cm.num_classes)
        validate_weights_size(weights.size()[0], cm.num_classes)

    # Increase floating point precision
    cm = cm.type(torch.float64)
    dice = 2 * (cm.diag() * weights) / (((cm.sum(dim=1) + cm.sum(dim=0)) * weights) + EPSILON)

    if ignore_index is not None:
        def remove_index(dice_vector):
            assert ignore_index <= len(dice_vector), "ignore_index {} is larger than the length " \
                                                     "of dice vector {}".format(ignore_index, len(dice_vector))
            indices = [x for x in range(len(dice_vector)) if x != ignore_index]
            return dice_vector[indices]

        return MetricsLambda(remove_index, dice)
    else:
        return dice


def compute_mean_dice_coefficient(cm: ConfusionMatrix, ignore_index: int = None):
    """
    Computes the mean Sørensen–Dice Coefficient.

    Args:
        cm (:obj:`ignite.metrics.ConfusionMatrix`): A confusion matrix representing the classification of data.
        ignore_index (int): An index of a class to ignore for computation.

    Returns:
        float: The mean Sørensen–Dice Coefficient.
    """
    return compute_dice_coefficient(cm=cm, ignore_index=ignore_index).mean()


def compute_mean_generalized_dice_coefficient(cm: ConfusionMatrix, weights: torch.Tensor, ignore_index: int = None):
    """
    Computes the mean Generalized Dice Coefficient.

    Args:
        cm (:obj:`ignite.metrics.ConfusionMatrix`): A confusion matrix representing the classification of data.
        weights (:obj:`torch.Tensor`): A tensor representing weights for each classes.
        ignore_index (int): An index of a class to ignore for computation.

    Returns:
        float: The mean Generalized Dice Coefficient.
    """
    return compute_generalized_dice_coefficient(cm=cm, ignore_index=ignore_index, weights=weights).mean()


def validate_ignore_index(ignore_index: int) -> None:
    """
    Check whether the ignore index is valid or not.

    Args:
        ignore_index (int): An index of a class to ignore for computation.
    """
    assert ignore_index >= 0, "ignore_index must be non-negative, but given {}".format(ignore_index)


def validate_num_classes(ignore_index: int, num_classes: int) -> None:
    """
    Check whether the num_classes is valid or not.

    Args:
        ignore_index (int): An index of a class to ignore for computation.
        num_classes (int): The number of classes in the problem (number of rows in the
    """
    assert ignore_index < num_classes, "ignore index must be lower than the number of classes in confusion matrix, " \
                                       "but {} was given".format(ignore_index)


def validate_weights_size(weights_size: int, num_classes: int) -> None:
    """
    Check whether if the size of given weights matches the number of classes of the problem.

    Args:
        weights_size (int): The size of a weight vector used in the loss computation.
        num_classes (int): The number of classes of the problem.
    """
    assert weights_size == num_classes, "Weights vector must be the same length than the number of " \
                                        "classes, but a size of {} was given".format(weights_size)



# -*- coding: utf-8 -*-
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import torch.nn

from typing import Union
from ignite.metrics.confusion_matrix import ConfusionMatrix
from ignite.metrics import MetricsLambda
from torch.autograd import Variable

from samitorch.metrics.metrics import validate_ignore_index
from samitorch.utils.utils import flatten, to_onehot

SUPPORTED_REDUCTIONS = [None, "mean"]

EPSILON = 1e-15


class DiceLoss(torch.nn.Module):
    """
    The Sørensen-Dice Loss.
    """

    def __init__(self, reduction: Union[None, str] = "mean"):
        super(DiceLoss, self).__init__()
        assert reduction in SUPPORTED_REDUCTIONS, "Reduction type not supported."
        self._reduction = reduction

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor, weights: torch.Tensor = None,
                ignore_index: int = None):
        """
        Computes the Sørensen–Dice loss.

        Note that PyTorch optimizers minimize a loss. In this case, we would like to maximize the dice loss so we
        return the negated dice loss.

        Args:
            inputs (:obj:`torch.Tensor`) : A tensor of shape (B, C, ..). The model prediction on which the loss has to be computed.
            targets (:obj:`torch.Tensor`) : A tensor of shape (B, C, ..). The ground truth.
            ignore_index (int): An index to ignore for computation.

        Returns:
            :obj:`torch.Tensor`: The Sørensen–Dice loss for each class or reduced according to reduction method.
        """
        if ignore_index is not None:
            validate_ignore_index(ignore_index)

        assert inputs.size() == targets.size(), "'Inputs' and 'Targets' must have the same shape."

        inputs = flatten(inputs)
        targets = flatten(targets)

        targets = targets.float()

        # Compute per channel Dice Coefficient
        intersect = (inputs * targets).sum(-1)

        if weights is not None:
            if weights.requires_grad is not False:
                weights.requires_grad = False
            intersect = weights * intersect

        denominator = (inputs + targets).sum(-1)

        dice = 1.0 - (2.0 * intersect / denominator.clamp(min=EPSILON))

        if ignore_index is not None:

            def ignore_index_fn(dice_vector):
                indices = list(range(len(dice_vector)))
                indices.remove(ignore_index)
                return dice_vector[indices]

            return MetricsLambda(ignore_index_fn, dice).compute()

        else:
            if self._reduction == "mean":
                dice = 1.0 - (2.0 * intersect / denominator.clamp(min=EPSILON)).mean()
            elif self._reduction is None:
                pass
            else:
                raise NotImplementedError("Reduction method not implemented.")
            return dice


class GeneralizedDiceLoss(torch.nn.Module):
    """
      The Generalized Sørensen-Dice Loss.
      """

    def __init__(self, reduction: Union[None, str] = "mean"):
        super(GeneralizedDiceLoss, self).__init__()
        assert reduction in SUPPORTED_REDUCTIONS, "Reduction type not supported."
        self._reduction = reduction

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor, ignore_index: int = None):
        """
        Computes the Sørensen–Dice loss.

        Note that PyTorch optimizers minimize a loss. In this case, we would like to maximize the dice loss so we
        return the negated dice loss.

        Args:
            inputs (:obj:`torch.Tensor`) : A tensor of shape (B, C, ..). The model prediction on which the loss has to be computed.
            targets (:obj:`torch.Tensor`) : A tensor of shape (B, C, ..). The ground truth.
            ignore_index (int): An index to ignore for computation.

        Returns:
            :obj:`torch.Tensor`: The Sørensen–Dice loss for each class or reduced according to reduction method.
        """
        if ignore_index is not None:
            validate_ignore_index(ignore_index)

        assert inputs.size() == targets.size(), "'Inputs' and 'Targets' must have the same shape."

        inputs = flatten(inputs)
        targets = flatten(targets)

        targets = targets.float()

        class_weights = torch.tensor(1.0 / torch.pow(targets.sum(-1), 2).clamp(min=1e-15), requires_grad=False,
                                     dtype=torch.float)

        # Compute per channel Dice Coefficient
        intersect = (inputs * targets).sum(-1) * class_weights

        denominator = (inputs + targets).sum(-1) * class_weights

        dice = 1.0 - (2.0 * intersect / denominator.clamp(min=EPSILON))

        if ignore_index is not None:

            def ignore_index_fn(dice_vector):
                indices = list(range(len(dice_vector)))
                indices.remove(ignore_index)
                return dice_vector[indices]

            return MetricsLambda(ignore_index_fn, dice).compute()

        else:
            if self._reduction == "mean":
                dice = 1.0 - (2.0 * intersect / denominator.clamp(min=EPSILON)).mean()
            elif self._reduction is None:
                pass
            else:
                raise NotImplementedError("Reduction method not implemented.")
            return dice


class WeightedCrossEntropyLoss(torch.nn.Module):

    def __init__(self, reduction="mean"):
        super(WeightedCrossEntropyLoss, self).__init__()
        assert reduction in SUPPORTED_REDUCTIONS, "Reduction is not supported."
        self._reduction = reduction

    def _validate_ignore_index(self, ignore_index: int, input_shape: int):
        """
        Validate the `ignore_index` variable.

        Args:
            ignore_index (int): An index of a class to ignore for computation.
            input_shape (int): Input tensor.
        """
        assert ignore_index >= 0, "ignore_index must be non-negative, but given {}".format(ignore_index)
        assert ignore_index < input_shape, "ignore index must be lower than the number of classes in " \
                                           "confusion matrix, but {} was given.".format(ignore_index)

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor, ignore_index: int = None) -> float:
        """
        Computes the Weighted Cross Entropy Loss (WCE) as described in https://arxiv.org/pdf/1707.03237.pdf

        Args:
            inputs (:obj:`torch.Tensor`): A tensor of shape (B, C, ..). The model's prediction on which the loss has to be computed.
            targets (:obj:`torch.Tensor`): A tensor of shape (B, C, ..). The ground truth.
            ignore_index (int): An index to ignore for computation.

       Returns:
           float: the weighted Cross-Entropy loss value.

        """
        if ignore_index is not None:
            self._validate_ignore_index(ignore_index, inputs.shape[1])
        else:
            ignore_index = -100

        class_weights = self._compute_class_weights(inputs)

        return torch.nn.functional.cross_entropy(inputs, targets, weight=class_weights,
                                                 ignore_index=ignore_index)

    @staticmethod
    def _compute_class_weights(inputs: torch.Tensor):
        """
        Compute weights for each class as described in https://arxiv.org/pdf/1707.03237.pdf

        Args:
            inputs: (:obj:`torch.Tensor`): A tensor of shape (B, C, ..). The model's prediction on which the loss has to be computed.

        Returns:
            :obj:`torch.Variable`: A variable containing class weights.
        """
        flattened_inputs = flatten(inputs)
        class_weights = torch.tensor((flattened_inputs.shape[1] - flattened_inputs.sum(-1)) / flattened_inputs.sum(-1),
                                     requires_grad=False, dtype=torch.float)
        return class_weights



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import unittest
import torch

from enum import Enum

from hamcrest import *

from samitorch.factories.factories import PaddingLayerFactory, ActivationLayerFactory, PoolingLayerFactory, \
    NormalizationLayerFactory
from samitorch.factories.enums import ActivationLayers, PoolingLayers, PaddingLayers, NormalizationLayers


class IncorrectLayers(Enum):
    Incorrect = "Incorrect"


class PaddingFactoryTest(unittest.TestCase):
    CORRECT_PADDING_STRATEGY = PaddingLayers.ReplicationPad3d
    INCORRECT_PADDING_STRATEGY = IncorrectLayers.Incorrect

    def setUp(self):
        self.factory = PaddingLayerFactory()

    def test_should_instantiate_padding_layer(self):
        padding_layer = self.factory.create_layer(self.CORRECT_PADDING_STRATEGY, dims=(1, 1, 1, 1, 1, 1))
        assert padding_layer is not None

    def test_should_instantiate_correct_padding_layer(self):
        padding_layer = self.factory.create_layer(self.CORRECT_PADDING_STRATEGY, dims=(1, 1, 1, 1, 1, 1))
        assert_that(padding_layer, instance_of(torch.nn.ReplicationPad3d))

    def test_should_raise_key_error_exception_with_unknown_padding_strategy(self):
        assert_that(
            calling(self.factory.create_layer).with_args(self.INCORRECT_PADDING_STRATEGY, dims=(1, 1, 1, 1, 1, 1)),
            raises(KeyError))

    def test_registering_new_padding_strategy_should_append_to_supported_strategies(self):
        new_padding_strategy = "ReplicationPad1d"
        creator = torch.nn.ReplicationPad1d
        self.factory.register(new_padding_strategy, creator)
        assert_that(self.factory._padding_strategies, has_entry(new_padding_strategy, creator))


class ActivationFunctionFactory(unittest.TestCase):
    CORRECT_ACTIVATION_FUNCTION = ActivationLayers.ReLU

    def setUp(self):
        self.factory = ActivationLayerFactory()

    def test_should_instantiate_activation_layer(self):
        activation_function = self.factory.create_layer(self.CORRECT_ACTIVATION_FUNCTION, inplace=True)
        assert activation_function is not None

    def test_should_instantiate_correct_activation_layer(self):
        activation_function = self.factory.create_layer(self.CORRECT_ACTIVATION_FUNCTION, inplace=True)
        assert_that(activation_function, instance_of(torch.nn.ReLU))

    def test_should_raise_key_error_exception_with_unknown_activation_function(self):
        assert_that(
            calling(self.factory.create_layer).with_args(IncorrectLayers.Incorrect),
            raises(KeyError))

    def test_registering_new_activation_function_should_append_to_supported_function(self):
        new_activation_function = "ELU"
        creator = torch.nn.ELU
        self.factory.register(new_activation_function, creator)
        assert_that(self.factory._activation_functions, has_entry(new_activation_function, creator))


class PoolingFactoryTest(unittest.TestCase):
    CORRECT_POOLING_STRATEGY = PoolingLayers.MaxPool3d
    INCORRECT_POOLING_STRATEGY = IncorrectLayers.Incorrect
    KERNEL_SIZE = 2

    def setUp(self):
        self.factory = PoolingLayerFactory()

    def test_should_instantiate_pooling_layer(self):
        pooling_layer = self.factory.create_layer(self.CORRECT_POOLING_STRATEGY, kernel_size=self.KERNEL_SIZE)
        assert pooling_layer is not None

    def test_should_instantiate_correct_pooling_layer(self):
        pooling_layer = self.factory.create_layer(self.CORRECT_POOLING_STRATEGY, kernel_size=self.KERNEL_SIZE)
        assert_that(pooling_layer, instance_of(torch.nn.MaxPool3d))

    def test_should_raise_key_error_exception_with_unknown_pooling_strategy(self):
        assert_that(
            calling(self.factory.create_layer).with_args(self.INCORRECT_POOLING_STRATEGY, kernel_size=self.KERNEL_SIZE),
            raises(KeyError))

    def test_should_instantiate_Conv3d_pooling_layer(self):
        pooling_layer = self.factory.create_layer(PoolingLayers.Conv3d, 2, 2, 16, 32)
        assert_that(pooling_layer, instance_of(torch.nn.Conv3d))
        assert_that(pooling_layer, has_property("stride", (2, 2, 2)))
        assert_that(pooling_layer, has_property("in_channels", 16))
        assert_that(pooling_layer, has_property("out_channels", 32))

    def test_registering_new_pooling_strategy_should_append_to_supported_strategies(self):
        new_pooling_strategy = "MaxPool1d"
        creator = torch.nn.MaxPool1d
        self.factory.register(new_pooling_strategy, creator)
        assert_that(self.factory._pooling_strategies, has_entry(new_pooling_strategy, creator))


class NormalizationFactoryTest(unittest.TestCase):
    GROUP_NORM = NormalizationLayers.GroupNorm
    BATCH_NORM = NormalizationLayers.BatchNorm3d
    INCORRECT_NORM_LAYER = IncorrectLayers.Incorrect
    NUM_FEATURES = 256
    NUM_GROUPS = 8

    def setUp(self):
        self.factory = NormalizationLayerFactory()

    def test_should_instantiate_normalization_layer(self):
        normalization_layer = self.factory.create_layer(self.GROUP_NORM, self.NUM_GROUPS,
                                                        self.NUM_FEATURES)
        assert normalization_layer is not None

    def test_should_instantiate_group_normalization(self):
        normalization_layer = self.factory.create_layer(self.GROUP_NORM, self.NUM_GROUPS,
                                                        self.NUM_FEATURES)
        assert_that(normalization_layer, instance_of(torch.nn.GroupNorm))

    def test_should_instantiate_batch_normalization(self):
        normalization_layer = self.factory.create_layer(self.BATCH_NORM, self.NUM_FEATURES)
        assert_that(normalization_layer, instance_of(torch.nn.BatchNorm3d))
        assert_that(normalization_layer, has_property("num_features", 256))

    def test_should_raise_key_error_exception_with_unknown_normalization_strategy(self):
        assert_that(
            calling(self.factory.create_layer).with_args(self.INCORRECT_NORM_LAYER, self.NUM_FEATURES),
            raises(KeyError))

    def test_registering_new_padding_strategy_should_append_to_supported_strategies(self):
        new_normalization_layer = "BatchNorm2d"
        creator = torch.nn.BatchNorm2d
        self.factory.register(new_normalization_layer, creator)
        assert_that(self.factory._normalization_strategies, has_entry(new_normalization_layer, creator))



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import abc
import torch

from samitorch.utils.model_io import save
from samitorch.metrics.gauges import RunningAverageGauge
from samitorch.inputs.batch import Batch
from samitorch.logger.plots import AccuracyPlot, LossPlot, ParameterPlot


class ModelTrainer(object):

    def __init__(self, config, callbacks, class_name):
        self._config = config
        self._callbacks = callbacks

        self._visdom = config.visdom

        self.training_metric_gauge = RunningAverageGauge()
        self.training_loss_gauge = RunningAverageGauge()
        self.validation_metric_gauge = RunningAverageGauge()
        self.validation_loss_gauge = RunningAverageGauge()

        self._training_metric_plot = AccuracyPlot(self._visdom,
                                                  "{} Training metric lr={} momentum={}".format(
                                                      class_name,
                                                      config.optimizer.param_groups[0]["lr"],
                                                      config.optimizer.param_groups[0]["momentum"]))
        self._validation_metric_plot = AccuracyPlot(self._visdom,
                                                    "{} Validation metric lr={} momentum={}".format(
                                                        class_name,
                                                        config.optimizer.param_groups[0]["lr"],
                                                        config.optimizer.param_groups[0]["momentum"]))
        self._training_loss_plot = LossPlot(self._visdom, "{} Training loss".format(class_name))
        self._validation_loss_plot = LossPlot(self._visdom, "{} Validation loss".format(class_name))
        self._learning_rate_plot = ParameterPlot(self._visdom, "{} Learning rate".format(class_name),
                                                 "learning rate")

        self._setup()

        self._global_step = torch.Tensor().new_zeros((1,), dtype=torch.int64, device='cpu')
        self._epoch = torch.Tensor().new_zeros((1,), dtype=torch.int64, device='cpu')

    @property
    def global_step(self):
        return self._global_step

    @property
    def epoch(self):
        return self._epoch

    @epoch.setter
    def epoch(self, epoch):
        self._epoch = epoch

    @property
    def config(self):
        return self._config

    def predict(self, batch: Batch, detach: bool = False):
        transformed_batch = Batch.from_batch(batch).to_device(self._config.running_config.device)
        if detach:
            transformed_batch.x = self._config.model(batch.x).detach()
        else:
            transformed_batch.x = self._config.model(batch.x)
        return transformed_batch

    def compute_metric(self):
        metric = self._config.metric.compute()
        self._config.metric.reset()
        return metric

    def evaluate_loss(self, X, y):
        return self._config.criterion(X, y)

    def update_lr(self, new_lr):
        self._config.optimizer.param_groups["lr"] = new_lr

    def update_metric_gauge(self, metric, n_data, phase="training"):
        if phase == "training":
            self.training_metric_gauge.update(metric, n_data)
        else:
            self.validation_metric_gauge.update(metric, n_data)

    def update_loss_gauge(self, loss, n_data, phase="training"):
        if phase == "training":
            self.training_loss_gauge.update(loss, n_data)
        else:
            self.validation_loss_gauge.update(loss, n_data)

    def update_metric_plot(self, step, value, phase="training"):
        if phase == "training":
            self._training_metric_plot.append(step, value)
        else:
            self._validation_metric_plot.append(step, value)

    def update_loss_plot(self, step, value, phase="training"):
        if phase == "training":
            self._training_loss_plot.append(step, value)
        else:
            self._validation_loss_plot.append(step, value)

    def update_learning_rate_plot(self, epoch, value):
        self._learning_rate_plot.append(epoch, value)

    def update_metric(self, pred, y):
        self._config.metric.update((pred, y))

    def step(self):
        self._config.optimizer.step()

    def at_training_begin(self):
        self._initialize_model(self._config.model)

    def at_training_end(self, epoch_num: int):
        save("model.pickle", self._config.model, epoch_num, self._config.optimizer)

    def at_epoch_begin(self):
        self._config.model.train()

    def at_epoch_end(self):
        self._epoch += 1
        self.training_metric_gauge.reset()
        self.training_loss_gauge.reset()

    def at_iteration_begin(self):
        self._config.optimizer.zero_grad()

    def at_iteration_end(self):
        self._global_step += 1
        self.enable_gradients()

    def at_validation_begin(self):
        self._config.model.eval()

    def at_validation_end(self):
        self.validation_metric_gauge.reset()
        self.validation_loss_gauge.reset()

    @abc.abstractmethod
    def finalize(self):
        raise NotImplementedError()

    def disable_gradients(self):
        for p in self._config.model.parameters(): p.requires_grad = False

    def enable_gradients(self):
        for p in self._config.model.parameters(): p.requires_grad = True

    def reduce_tensor(self, tensor):
        rt = tensor.clone()
        torch.distributed.all_reduce(rt, op=torch.distributed.ReduceOp.SUM)
        rt /= int(self._config.running_config.world_size)
        return rt

    def reset_optimizer(self):
        self._config.optimizer.zero_grad()

    @staticmethod
    def _initialize_model(model):
        for m in model.modules():
            if isinstance(m, torch.nn.Conv3d):
                torch.nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')

    @abc.abstractmethod
    def _setup(self):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import unittest
import os
import nibabel as nib
import numpy as np
import collections
import torch

from torchvision.transforms import transforms

from hamcrest import *

from samitorch.inputs.sample import Sample

from samitorch.inputs.transformers import LoadNifti, ToNifti1Image, RemapClassIDs, ApplyMask, \
    ToNumpyArray, RandomCrop, NiftiToDisk, To2DNifti1Image, ToPNGFile, RandomCrop3D, \
    NiftiImageToNumpy, ResampleNiftiImageToTemplate, CropToContent, ToTensorPatches, ToNDTensor, ToNDArrayPatches


class ToNumpyArrayTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1.nii")
    VALID_3D_LABELS = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")
    INVALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "invalid_T1.nii")
    INVALID_3D_LABELS = os.path.join(TEST_DATA_FOLDER_PATH, "invalid_mask.nii")

    def setUp(self):
        pass

    def test_should_return_a_sample_containing_both_Numpy_ndarrays(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE, y=self.VALID_3D_LABELS, is_labeled=True)

        transformed_sample = ToNumpyArray().__call__(sample)

        assert_that(transformed_sample.x, instance_of(np.ndarray))
        assert_that(transformed_sample.y, instance_of(np.ndarray))

    def test_should_return_a_single_Numpy_ndarray_from_sample(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE)
        transformed_sample = ToNumpyArray().__call__(sample)

        assert_that(transformed_sample.x, instance_of(np.ndarray))
        assert_that(transformed_sample.y, is_(None))

    def test_should_return_a_single_Numpy_ndarray_from_string(self):
        nd_array = ToNumpyArray().__call__(self.VALID_3D_LABELS)
        assert_that(nd_array, instance_of(np.ndarray))

    def test_should_return_a_single_Numpy_ndarray_from_Nifti1Image(self):
        nifti_image = nib.load(self.VALID_3D_LABELS)

        nd_array = ToNumpyArray().__call__(nifti_image)
        assert_that(nd_array, instance_of(np.ndarray))

    def test_should_raise_FileNotFound_with_x_invalid_type(self):
        sample = Sample(x=self.INVALID_3D_NIFTI_FILE)

        assert_that(calling(ToNumpyArray()).with_args(sample), raises(FileNotFoundError))

    def test_should_raise_FileNotFound_with_y_invalid_type(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE, y=self.INVALID_3D_LABELS, is_labeled=True)

        assert_that(calling(ToNumpyArray()).with_args(sample), raises(FileNotFoundError))

    def test_should_raise_FileNotFound_with_two_invalid_path(self):
        sample = Sample(x=self.INVALID_3D_NIFTI_FILE, y=self.INVALID_3D_LABELS, is_labeled=True)

        assert_that(calling(ToNumpyArray()).with_args(sample), raises(FileNotFoundError))


class LoadNiftiTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1.nii")
    VALID_3D_LABELS = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")
    VALID_4D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "FA.nii")
    INVALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "invalid_T1.nii")
    INVVALID_3D_LABELS = os.path.join(TEST_DATA_FOLDER_PATH, "invalid_mask.nii")

    def setUp(self):
        pass

    def test_should_return_a_sample_containing_both_Nifti1Image(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE, y=self.VALID_3D_LABELS, is_labeled=True)

        transformed_sample = LoadNifti().__call__(sample)

        assert_that(transformed_sample.x, instance_of(nib.Nifti1Image))
        assert_that(transformed_sample.y, instance_of(nib.Nifti1Image))

    def test_should_return_a_single_Nifti1Image(self):
        nifti1Image = LoadNifti().__call__(self.VALID_3D_NIFTI_FILE)

        assert_that(nifti1Image, instance_of(nib.Nifti1Image))

    def test_should_return_a_sample_containing_one_Nifti1Image(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE)
        transformed_sample = LoadNifti().__call__(sample)

        assert_that(transformed_sample.x, instance_of(nib.Nifti1Image))
        assert_that(transformed_sample.y, is_(None))

    def test_should_raise_FileNotFoundError_with_x_invalid_path(self):
        sample = Sample(x=self.INVALID_3D_NIFTI_FILE)

        assert_that(calling(LoadNifti()).with_args(sample), raises(FileNotFoundError))

    def test_should_raise_FileNotFoundError_with_y_invalid_path(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE, y=self.INVVALID_3D_LABELS, is_labeled=True)

        assert_that(calling(LoadNifti()).with_args(sample), raises(FileNotFoundError))

    def test_should_raise_FileNotFoundError_with_two_invalid_path(self):
        sample = Sample(x=self.INVALID_3D_NIFTI_FILE, y=self.INVVALID_3D_LABELS, is_labeled=True)

        assert_that(calling(LoadNifti()).with_args(sample), raises(FileNotFoundError))

    def test_should_raise_FileNotFoundError_with_one_invalid_path(self):
        assert_that(calling(LoadNifti()).with_args(self.INVALID_3D_NIFTI_FILE), raises(FileNotFoundError))

    def test_should_return_nifti_image_with_valid_4d_nifti_file(self):
        nifti_4d_image = LoadNifti().__call__(self.VALID_4D_NIFTI_FILE)
        assert_that(nifti_4d_image, is_not(None))
        assert_that(nifti_4d_image, is_(nib.Nifti1Image))


class NiftiImageToNumpyTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1.nii")
    VALID_3D_LABELS = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")
    INVALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "invalid_T1.nii")
    INVVALID_3D_LABELS = os.path.join(TEST_DATA_FOLDER_PATH, "invalid_mask.nii")

    def setUp(self):
        pass

    def test_should_return_a_sample_containing_both_Numpy_ndarrays(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE, y=self.VALID_3D_LABELS, is_labeled=True)

        transformed_sample = LoadNifti().__call__(sample)
        transformed_sample = NiftiImageToNumpy().__call__(transformed_sample)

        assert_that(transformed_sample.x, instance_of(np.ndarray))
        assert_that(transformed_sample.y, instance_of(np.ndarray))

    def test_should_return_a_single_Numpy_ndarray(self):
        nifti1Image = LoadNifti().__call__(self.VALID_3D_NIFTI_FILE)
        nd_array = NiftiImageToNumpy().__call__(nifti1Image)

        assert_that(nd_array, instance_of(np.ndarray))

    def test_should_return_a_sample_containing_one_Numpy_ndarray(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE)
        transformed_sample = LoadNifti().__call__(sample)
        transformed_sample = NiftiImageToNumpy().__call__(transformed_sample)

        assert_that(transformed_sample.x, instance_of(np.ndarray))
        assert_that(transformed_sample.y, is_(None))

    def test_should_raise_TypeError_with_x_invalid_type(self):
        sample = Sample(x=self.INVALID_3D_NIFTI_FILE)

        assert_that(calling(NiftiImageToNumpy()).with_args(sample), raises(TypeError))

    def test_should_raise_TypeError_with_y_invalid_type(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE, y=self.INVVALID_3D_LABELS, is_labeled=True)

        assert_that(calling(NiftiImageToNumpy()).with_args(sample), raises(TypeError))

    def test_should_raise_TypeError_with_two_invalid_path(self):
        sample = Sample(x=self.INVALID_3D_NIFTI_FILE, y=self.INVVALID_3D_LABELS, is_labeled=True)

        assert_that(calling(NiftiImageToNumpy()).with_args(sample), raises(TypeError))

    def test_should_raise_TypeError_with_one_invalid_type(self):
        assert_that(calling(NiftiImageToNumpy()).with_args(self.INVALID_3D_NIFTI_FILE), raises(TypeError))


class ResampleNiftiImageToTemplateTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/resample")
    OUTPUT_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/generated/ResampleNiftiImageToTemplate")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1.nii")
    VALID_3D_1MM_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1_1mm.nii")
    INVALID_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "invalid.nii")
    RESAMPLED_NIFTI_FROM_NIFTI = os.path.join(OUTPUT_DATA_FOLDER_PATH, "resampled_nifti_from_nifti.nii")
    RESAMPLED_NIFTI_FROM_STRING = os.path.join(OUTPUT_DATA_FOLDER_PATH, "resampled_nifti_from_string.nii")
    RESAMPLED_NIFTI_FROM_STRING_TEMPLATE = os.path.join(OUTPUT_DATA_FOLDER_PATH,
                                                        "resampled_nifti_from_string_template.nii")

    ALL = [RESAMPLED_NIFTI_FROM_NIFTI, RESAMPLED_NIFTI_FROM_STRING, RESAMPLED_NIFTI_FROM_STRING_TEMPLATE]

    @classmethod
    def _flatten(cls, l):
        for el in l:
            if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
                yield from cls._flatten(el)
            else:
                yield el

    def setUp(self):
        pass

    @classmethod
    def setUpClass(cls):
        all_files = cls._flatten(cls.ALL)
        for element in all_files:
            if os.path.exists(element):
                os.remove(element)

        if not os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH):
            os.makedirs(cls.OUTPUT_DATA_FOLDER_PATH)

        assert_that(len(os.listdir(cls.OUTPUT_DATA_FOLDER_PATH)), is_(0))
        assert_that(os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH), is_(True))

    def test_should_return_resampled_Nifti1Image_from_Nifti1Image(self):
        transform_ = ResampleNiftiImageToTemplate(interpolation="continuous",
                                                  clip=False,
                                                  template=nib.load(self.VALID_3D_1MM_NIFTI_FILE))
        nifti_image = transform_(nib.load(self.VALID_3D_NIFTI_FILE))

        np.testing.assert_array_equal([256, 256, 192], nifti_image.shape)
        assert_that(nifti_image, instance_of(nib.Nifti1Image))
        transform_ = NiftiToDisk(self.RESAMPLED_NIFTI_FROM_NIFTI)
        transform_(nifti_image)

    def test_should_return_resampled_Nifti1Image_from_string(self):
        transform_ = ResampleNiftiImageToTemplate(interpolation="continuous",
                                                  clip=False,
                                                  template=nib.load(self.VALID_3D_1MM_NIFTI_FILE))
        nifti_image = transform_(self.VALID_3D_NIFTI_FILE)
        np.testing.assert_array_equal([256, 256, 192], nifti_image.shape)
        assert_that(nifti_image, instance_of(nib.Nifti1Image))
        transform_ = NiftiToDisk(self.RESAMPLED_NIFTI_FROM_STRING)
        transform_(nifti_image)

    def test_should_return_resampled_Nifti1Image_from_string_template(self):
        transform_ = ResampleNiftiImageToTemplate(interpolation="continuous",
                                                  clip=False,
                                                  template=self.VALID_3D_1MM_NIFTI_FILE)
        nifti_image = transform_(self.VALID_3D_NIFTI_FILE)
        np.testing.assert_array_equal([256, 256, 192], nifti_image.shape)
        assert_that(nifti_image, instance_of(nib.Nifti1Image))
        transform_ = NiftiToDisk(self.RESAMPLED_NIFTI_FROM_STRING_TEMPLATE)
        transform_(nifti_image)

    def test_should_return_resampled_Nifti1Image_from_Nifti1Image_from_string_template(self):
        transform_ = ResampleNiftiImageToTemplate(interpolation="continuous",
                                                  clip=False,
                                                  template=self.VALID_3D_1MM_NIFTI_FILE)
        nifti_image = transform_(nib.load(self.VALID_3D_NIFTI_FILE))
        np.testing.assert_array_equal([256, 256, 192], nifti_image.shape)
        assert_that(nifti_image, instance_of(nib.Nifti1Image))
        transform_ = NiftiToDisk(self.RESAMPLED_NIFTI_FROM_STRING_TEMPLATE)
        transform_(nifti_image)

    def test_should_return_resampled_Nifti1Image_from_sample_Nifti1Image(self):
        sample = Sample(x=nib.load(self.VALID_3D_NIFTI_FILE), template=nib.load(self.VALID_3D_1MM_NIFTI_FILE))
        transform_ = ResampleNiftiImageToTemplate(interpolation="continuous",
                                                  clip=False)
        transformed_sample = transform_(sample)
        np.testing.assert_array_equal([256, 256, 192], transformed_sample.x.shape)
        assert_that(transformed_sample.x, instance_of(nib.Nifti1Image))


class ToNifti1ImageTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1.nii")
    VALID_3D_LABELS = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")

    def setUp(self):
        pass

    def test_should_return_a_Nifti1Image_from_Numpy_ndarray(self):
        header = nib.load(self.VALID_3D_LABELS).header
        nd_array = ToNumpyArray().__call__(self.VALID_3D_LABELS)
        transform_ = ToNifti1Image(header)
        new_nifti_image = transform_(nd_array)
        assert_that(new_nifti_image, instance_of(nib.nifti1.Nifti1Image))

    def test_should_return_a_Nifti1Image_Sample_from_Sample(self):
        header_image = nib.load(self.VALID_3D_LABELS).header
        header_labels = nib.load(self.VALID_3D_LABELS).header
        sample = Sample(x=self.VALID_3D_LABELS, y=self.VALID_3D_NIFTI_FILE, is_labeled=True)
        nd_array_sample = ToNumpyArray().__call__(sample)
        transform_ = ToNifti1Image([header_image, header_labels])
        nifti_image_sample = transform_(nd_array_sample)

        assert_that(nifti_image_sample.x, instance_of(nib.Nifti1Image))
        assert_that(nifti_image_sample.y, instance_of(nib.Nifti1Image))

    def test_should_return_a_Nifti1Image_Sample_from_single_element_Sample(self):
        header_image = nib.load(self.VALID_3D_LABELS).header
        sample = Sample(x=self.VALID_3D_LABELS)
        nd_array_sample = ToNumpyArray().__call__(sample)
        transform_ = ToNifti1Image([header_image])
        nifti_image_sample = transform_(nd_array_sample)

        assert_that(nifti_image_sample.x, instance_of(nib.Nifti1Image))
        assert_that(nifti_image_sample.y, is_(None))

    def test_should_return_sample_from_single_element_Sample_with_one_None_Header(self):
        sample = Sample(x=self.VALID_3D_LABELS)
        nd_array_sample = ToNumpyArray().__call__(sample)
        transform_ = ToNifti1Image([None])
        transformed_sample = transform_(nd_array_sample)

        assert_that(transformed_sample.x, instance_of(nib.Nifti1Image))

    def test_should_return_a_Nifti1Image_Sample_from_Sample_without_header(self):
        sample = Sample(x=self.VALID_3D_LABELS, y=self.VALID_3D_NIFTI_FILE, is_labeled=True)
        nd_array_sample = ToNumpyArray().__call__(sample)
        transform_ = ToNifti1Image([None, None])
        nifti_image_sample = transform_(nd_array_sample)

        assert_that(nifti_image_sample.x, instance_of(nib.Nifti1Image))
        assert_that(nifti_image_sample.y, instance_of(nib.Nifti1Image))

    def test_should_raise_TypeErrorException_with_non_ndarray_sample(self):
        sample = Sample(x=self.VALID_3D_LABELS, y=self.VALID_3D_NIFTI_FILE, is_labeled=True)
        transform_ = ToNifti1Image([None])

        assert_that(calling(transform_).with_args(sample), raises(TypeError))

    def test_should_raise_TypeErrorException_with_incorrect_dims_ndarray_in_x_sample(self):
        sample = Sample(x=np.random.randint(0, 255, (32, 32), dtype=np.int16))
        transform_ = ToNifti1Image([None])

        assert_that(calling(transform_).with_args(sample), raises(TypeError))

    def test_should_raise_TypeErrorException_with_incorrect_dims_ndarray_in_y_sample(self):
        sample = Sample(x=np.random.randint(0, 255, (32, 32), dtype=np.int16),
                        y=np.random.randint(0, 255, (32, 32), dtype=np.int16), is_labeled=True)
        transform_ = ToNifti1Image([None, None])

        assert_that(calling(transform_).with_args(sample), raises(TypeError))

    def test_should_return_a_nifti1Image_sample_with_non_list_header_while_sample_is_labeled(self):
        header_image = nib.load(self.VALID_3D_LABELS).header
        sample = Sample(x=self.VALID_3D_LABELS, y=self.VALID_3D_NIFTI_FILE, is_labeled=True)
        nd_array_sample = ToNumpyArray().__call__(sample)
        transform_ = ToNifti1Image(header_image)

        assert_that(calling(transform_).with_args(nd_array_sample), raises(TypeError))

    def test_should_raise_TypeError_with_non_list_header_with_Sample(self):
        header_image = nib.load(self.VALID_3D_NIFTI_FILE).header
        header_labels = nib.load(self.VALID_3D_LABELS).header
        sample = Sample(x=self.VALID_3D_NIFTI_FILE, y=self.VALID_3D_LABELS, is_labeled=True)
        nd_array_sample = ToNumpyArray().__call__(sample)
        transform_ = ToNifti1Image(header_image)

        assert_that(calling(transform_).with_args(nd_array_sample), raises(TypeError))

    def test_should_raise_AssertionError_with_non_list_header_while_sample_has_bad_header(self):
        sample = Sample(x=self.VALID_3D_LABELS, y=self.VALID_3D_NIFTI_FILE, is_labeled=True)
        nd_array_sample = ToNumpyArray().__call__(sample)
        transform_ = ToNifti1Image(["bad_header", np.random.randint(0, 255, (32, 32))])

        assert_that(calling(transform_).with_args(nd_array_sample), raises(AttributeError))


class NiftiToDiskTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    OUTPUT_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/generated/NiftiToDisk")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1.nii")
    VALID_3D_LABELS = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")
    SAVED_NIFTI_FILE_FROM_NIFTI = os.path.join(OUTPUT_DATA_FOLDER_PATH, "T1_saved_from_nifti.nii")
    SAVED_NIFTI_FILE_FROM_SAMPLE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "T1_saved_from_sample.nii")
    SAVED_LABELS_FROM_SAMPLE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "Mask_saved_from_sample.nii")
    ALL = [SAVED_NIFTI_FILE_FROM_NIFTI, SAVED_NIFTI_FILE_FROM_SAMPLE, SAVED_LABELS_FROM_SAMPLE]

    @classmethod
    def _flatten(cls, l):
        for el in l:
            if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
                yield from cls._flatten(el)
            else:
                yield el

    def setUp(self):
        pass

    @classmethod
    def setUpClass(cls):
        all_files = cls._flatten(cls.ALL)
        for element in all_files:
            if os.path.exists(element):
                os.remove(element)

        if not os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH):
            os.makedirs(cls.OUTPUT_DATA_FOLDER_PATH)

        assert_that(len(os.listdir(cls.OUTPUT_DATA_FOLDER_PATH)), is_(0))
        assert_that(os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH), is_(True))

    def test_should_save_nifti_file_from_Nifti1Image(self):
        nifti_image = nib.load(self.VALID_3D_NIFTI_FILE)
        transform_ = NiftiToDisk(self.SAVED_NIFTI_FILE_FROM_NIFTI)
        transform_(nifti_image)

        assert_that(os.path.exists(self.SAVED_NIFTI_FILE_FROM_NIFTI))

    def test_should_save_nifti_file_from_sample(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE, y=self.VALID_3D_LABELS, is_labeled=True)
        header_image = nib.load(self.VALID_3D_NIFTI_FILE).header
        header_label = nib.load(self.VALID_3D_LABELS).header

        transformed_sample = ToNumpyArray().__call__(sample)
        transformed_sample = ToNifti1Image([header_image, header_label]).__call__(transformed_sample)
        transform_ = NiftiToDisk([self.SAVED_NIFTI_FILE_FROM_SAMPLE, self.SAVED_LABELS_FROM_SAMPLE])
        transform_(transformed_sample)

        assert_that(os.path.exists(self.SAVED_NIFTI_FILE_FROM_SAMPLE))
        assert_that(os.path.exists(self.SAVED_LABELS_FROM_SAMPLE))

    def test_should_raise_TypeError_exception_when_not_passing_a_list_of_file_paths(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE, y=self.VALID_3D_LABELS, is_labeled=True)
        header_image = nib.load(self.VALID_3D_NIFTI_FILE).header
        header_label = nib.load(self.VALID_3D_LABELS).header

        transformed_sample = ToNumpyArray().__call__(sample)
        transformed_sample = ToNifti1Image([header_image, header_label]).__call__(transformed_sample)
        transform_ = NiftiToDisk("bad_input")

        assert_that(calling(transform_).with_args(transformed_sample), raises(ValueError))


class ApplyMaskTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    OUTPUT_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/generated/ApplyMaskToNiftiImage")
    VALID_MASK_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")
    INVALID_MASK_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "invalid_mask.nii")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1.nii")
    MASKED_T1_FILE_FROM_SAMPLE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "masked_t1_file_from_sample.nii")
    MASK_FILE_FROM_SAMPLE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "mask.nii")

    ALL = [MASKED_T1_FILE_FROM_SAMPLE, MASK_FILE_FROM_SAMPLE]

    @classmethod
    def _flatten(cls, l):
        for el in l:
            if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
                yield from cls._flatten(el)
            else:
                yield el

    def setUp(self):
        pass

    @classmethod
    def setUpClass(cls):
        all_files = cls._flatten(cls.ALL)
        for element in all_files:
            if os.path.exists(element):
                os.remove(element)

        if not os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH):
            os.makedirs(cls.OUTPUT_DATA_FOLDER_PATH)

        assert_that(len(os.listdir(cls.OUTPUT_DATA_FOLDER_PATH)), is_(0))
        assert_that(os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH), is_(True))

    def test_transformer_should_initialize_with_str(self):
        transform_ = ApplyMask(self.VALID_MASK_FILE)
        assert_that(transform_, is_not(None))
        assert_that(transform_._mask, instance_of(np.ndarray))

    def test_transformer_should_initialize_with_Nifti1Image(self):
        transform_ = ApplyMask(nib.load(self.VALID_MASK_FILE))
        assert_that(transform_, is_not(None))
        assert_that(transform_._mask, instance_of(np.ndarray))

    def test_transformer_should_initialize_with_Numpy_ndarray(self):
        transform_ = ApplyMask(nib.load(self.VALID_MASK_FILE).get_fdata().__array__())
        assert_that(transform_, is_not(None))
        assert_that(transform_._mask, instance_of(np.ndarray))

    def test_should_initialize_mask_correctly(self):
        transformer = ApplyMask(self.VALID_MASK_FILE)
        np.testing.assert_array_less(transformer._mask, 1.0 + 1e-10)
        assert_that(len(transformer._mask.shape), is_(3))

    def test_returned_image_should_be_Nifti1Image_type(self):
        transformer = ApplyMask(self.VALID_MASK_FILE)
        nifti = nib.load(self.VALID_3D_NIFTI_FILE)
        transformed_nifti = transformer.__call__(nifti)
        assert_that(transformed_nifti, instance_of(nib.Nifti1Image))

    def test_should_return_image_with_same_shape_as_original(self):
        transformer = ApplyMask(self.VALID_MASK_FILE)
        nifti = nib.load(self.VALID_3D_NIFTI_FILE)
        transformed_nifti = transformer.__call__(nifti)
        np.testing.assert_array_equal(transformed_nifti.shape, nifti.shape)

    def test_should_return_sample_with_masked_image(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE, y=self.VALID_MASK_FILE, is_labeled=True)
        transformed_sample = LoadNifti().__call__(sample)
        transformed_sample = ApplyMask().__call__(transformed_sample)

        assert_that(transformed_sample.x, instance_of(nib.Nifti1Image))

    def test_should_raise_TypeError_exception_with_invalid_masked_image(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE, y=self.VALID_MASK_FILE, is_labeled=True)
        transformed_sample = LoadNifti().__call__(sample)
        transformed_sample.y = "incorrect_input"
        transform_ = ApplyMask()

        assert_that(calling(transform_).with_args(transformed_sample), raises(TypeError))

    def test_should_raise_TypeError_exception_with_invalid_argument(self):
        transform_ = ApplyMask()

        assert_that(calling(transform_).with_args("invalid_argument"), raises(TypeError))

    def test_should_return_sample_with_masked_image_for_inspection(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE, y=self.VALID_MASK_FILE, is_labeled=True)
        transformed_sample = LoadNifti().__call__(sample)
        transformed_sample = ApplyMask().__call__(transformed_sample)

        transform_ = NiftiToDisk([self.MASKED_T1_FILE_FROM_SAMPLE, self.MASK_FILE_FROM_SAMPLE])
        transform_(transformed_sample)

        assert_that(os.path.exists(self.MASKED_T1_FILE_FROM_SAMPLE))
        assert_that(os.path.exists(self.MASK_FILE_FROM_SAMPLE))

    def test_should_raise_FileNotFoundError_exception_with_invalid_file(self):
        assert_that(calling(ApplyMask).with_args(self.INVALID_MASK_FILE), raises(FileNotFoundError))

    def test_should_raise_TypeError_exception_when_passing_invalid_x_sample(self):
        sample = Sample(x="invalid", y=np.random.randint(0, 255, (32, 32)), is_labeled=True)
        transform_ = ApplyMask()

        assert_that(calling(transform_).with_args(sample), raises(TypeError))

    def test_shoud_raise_TypeError_exception_with_invalid_argument(self):
        transform_ = ApplyMask()

        assert_that(calling(transform_).with_args("invalid"), raises(TypeError))


class RemapClassIDsTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    OUTPUT_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/generated/RemapClassIDs")
    VALID_MASK_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1.nii")
    REMAPPED_LABELS_FROM_SAMPLE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "remapped_labels_from_sample.nii")
    UNTOUCHED_IMAGE_FROM_SAMPLE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "original_image_from_sample.nii")
    REMAPPED_LABELS_FROM_NDARRAY = os.path.join(OUTPUT_DATA_FOLDER_PATH, "remapped_labels_from_ndarray.nii")
    ALL = [UNTOUCHED_IMAGE_FROM_SAMPLE, REMAPPED_LABELS_FROM_SAMPLE, REMAPPED_LABELS_FROM_NDARRAY]

    @classmethod
    def _flatten(cls, l):
        for el in l:
            if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
                yield from cls._flatten(el)
            else:
                yield el

    def setUp(self):
        pass

    @classmethod
    def setUpClass(cls):
        all_files = cls._flatten(cls.ALL)
        for element in all_files:
            if os.path.exists(element):
                os.remove(element)

        if not os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH):
            os.makedirs(cls.OUTPUT_DATA_FOLDER_PATH)

        assert_that(len(os.listdir(cls.OUTPUT_DATA_FOLDER_PATH)), is_(0))
        assert_that(os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH), is_(True))

    def test_should_remap_labels_correctly_from_Numpy_ndarray(self):
        nd_array = ToNumpyArray().__call__(self.VALID_MASK_FILE)

        transform_ = RemapClassIDs([1, 2, 3], [4, 8, 12])

        remapped_nd_array = transform_(nd_array)

        expected_nd_array = nd_array * 4
        assert_that(np.isnan(np.all(remapped_nd_array)), is_(False))
        assert_that(np.isnan(np.all(remapped_nd_array)), is_(False))
        assert_that(np.isnan(np.all(expected_nd_array)), is_(False))
        assert_that(np.isnan(np.all(expected_nd_array)), is_(False))
        np.testing.assert_array_equal(remapped_nd_array, expected_nd_array)

    def test_should_fail_with_non_integer_class_ids(self):
        assert_that(calling(RemapClassIDs).with_args([1, 2, 3], [1.5, 4, 8]), raises(ValueError))

    def test_should_return_remapped_sample_from_Nifti1Image_sample(self):
        x_nifti = nib.load(self.VALID_3D_NIFTI_FILE)
        y_nifti = nib.load(self.VALID_MASK_FILE)
        sample = Sample(x=x_nifti, y=y_nifti, is_labeled=True)
        transform_ = RemapClassIDs([1, 2, 3], [4, 8, 12])

        remapped_sample = transform_(sample)

        expected_x = x_nifti
        expected_y = nib.Nifti1Image(y_nifti.get_fdata().__array__() * 4, affine=y_nifti.affine, header=y_nifti.header)

        np.testing.assert_array_equal(remapped_sample.y.get_fdata(), expected_y.get_fdata())
        np.testing.assert_array_equal(remapped_sample.x.get_fdata(), expected_x.get_fdata())

    def test_should_return_remapped_sample_from_Nifti1Image_for_inspection(self):
        x_nifti = nib.load(self.VALID_3D_NIFTI_FILE)
        y_nifti = nib.load(self.VALID_MASK_FILE)
        sample = Sample(x=x_nifti, y=y_nifti, is_labeled=True)
        transform_ = RemapClassIDs([1, 2, 3], [4, 8, 12])

        remapped_sample = transform_(sample)

        NiftiToDisk([self.UNTOUCHED_IMAGE_FROM_SAMPLE, self.REMAPPED_LABELS_FROM_SAMPLE]).__call__(remapped_sample)

    def test_should_return_remapped_sample_from_Numpy_ndarray_sample_for_inspection(self):
        x = ToNumpyArray().__call__(self.VALID_3D_NIFTI_FILE)
        y = ToNumpyArray().__call__(self.VALID_MASK_FILE)

        header_image = nib.load(self.VALID_3D_NIFTI_FILE).header
        header_label = nib.load(self.VALID_MASK_FILE).header
        sample = Sample(x=x, y=y, is_labeled=True)

        transform_ = RemapClassIDs([1, 2, 3], [4, 8, 12])

        remapped_sample = transform_(sample)

        nifti_sample = ToNifti1Image([header_image, header_label]).__call__(remapped_sample)

        NiftiToDisk([self.UNTOUCHED_IMAGE_FROM_SAMPLE, self.REMAPPED_LABELS_FROM_SAMPLE]).__call__(nifti_sample)

    def test_should_remap_labels_correctly_from_Numpy_ndarray_for_inspection(self):
        nd_array = ToNumpyArray().__call__(self.VALID_MASK_FILE)
        nifti_image = nib.load(self.VALID_MASK_FILE)
        transform_ = RemapClassIDs([1, 2, 3], [4, 8, 12])
        remapped_nd_array = transform_(nd_array)
        expected_nd_array = nd_array * 4

        assert_that(np.isnan(np.all(remapped_nd_array)), is_(False))
        assert_that(np.isnan(np.all(remapped_nd_array)), is_(False))
        assert_that(np.isnan(np.all(expected_nd_array)), is_(False))
        assert_that(np.isnan(np.all(expected_nd_array)), is_(False))
        np.testing.assert_array_equal(remapped_nd_array, expected_nd_array)

        new_nifti_image = nib.Nifti1Image(remapped_nd_array.transpose((3, 2, 1, 0)), nifti_image.affine,
                                          nifti_image.header)
        nib.save(new_nifti_image, self.REMAPPED_LABELS_FROM_NDARRAY)

    def test_should_raise_TypeError_exception_with_2D_Numpy_ndarray(self):
        nd_array = np.random.randint(0, 3, (32, 32))
        transform_ = RemapClassIDs([1, 2, 3], [4, 8, 12])
        remapped_nd_array = transform_(nd_array)
        expected_nd_array = nd_array * 4

        assert_that(np.isnan(np.all(remapped_nd_array)), is_(False))
        assert_that(np.isnan(np.all(remapped_nd_array)), is_(False))
        assert_that(np.isnan(np.all(expected_nd_array)), is_(False))
        assert_that(np.isnan(np.all(expected_nd_array)), is_(False))
        np.testing.assert_array_equal(remapped_nd_array, expected_nd_array)


class To2DNifti1ImageTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1.nii")
    VALID_MASK_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")
    OUTPUT_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/generated/To2DNifti1Image")
    NIFTI_2D_IMAGE_FROM_NDARRAY = os.path.join(OUTPUT_DATA_FOLDER_PATH, "2d_nifti_image_from_nd_array.nii")
    NIFTI_2D_IMAGE_FROM_SAMPLE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "2d_nifti_image_from_sample.nii")
    NIFTI_2D_LABEL_FROM_SAMPLE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "2d_nifti_label_from_sample.nii")
    ALL = [NIFTI_2D_IMAGE_FROM_NDARRAY, NIFTI_2D_IMAGE_FROM_SAMPLE, NIFTI_2D_LABEL_FROM_SAMPLE]

    @classmethod
    def _flatten(cls, l):
        for el in l:
            if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
                yield from cls._flatten(el)
            else:
                yield el

    def setUp(self):
        pass

    @classmethod
    def setUpClass(cls):
        all_files = cls._flatten(cls.ALL)
        for element in all_files:
            if os.path.exists(element):
                os.remove(element)

        if not os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH):
            os.makedirs(cls.OUTPUT_DATA_FOLDER_PATH)

        assert_that(len(os.listdir(cls.OUTPUT_DATA_FOLDER_PATH)), is_(0))
        assert_that(os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH), is_(True))

    def test_should_return_Nifti1Image_from_Numpy_array(self):
        nifti_image = nib.load(self.VALID_3D_NIFTI_FILE)
        nd_array = nib.load(self.VALID_3D_NIFTI_FILE).get_fdata().__array__()
        transform_ = To2DNifti1Image(nifti_image.header)
        nd_array_2d = transform_(nd_array)

        assert_that(nd_array_2d, instance_of(nib.Nifti1Image))

    def test_should_return_Sample_from_Sample_of_Numpy_array(self):
        header_x = nib.load(self.VALID_3D_NIFTI_FILE).header
        header_y = nib.load(self.VALID_MASK_FILE).header

        x = nib.load(self.VALID_3D_NIFTI_FILE).get_fdata().__array__()
        y = nib.load(self.VALID_MASK_FILE).get_fdata().__array__()

        sample = Sample(x=x, y=y, is_labeled=True)

        transform_ = To2DNifti1Image([header_x, header_y])
        transformed_sample = transform_(sample)

        assert_that(transformed_sample.x, instance_of(nib.Nifti1Image))
        assert_that(transformed_sample.y, instance_of(nib.Nifti1Image))

    def test_should_return_Sample_from_Sample_of_Numpy_array_for_inspection(self):
        header_x = nib.load(self.VALID_3D_NIFTI_FILE).header
        header_y = nib.load(self.VALID_MASK_FILE).header

        x = nib.load(self.VALID_3D_NIFTI_FILE).get_fdata().__array__()
        y = nib.load(self.VALID_MASK_FILE).get_fdata().__array__()

        sample = Sample(x=x, y=y, is_labeled=True)
        transform_ = RandomCrop(64)
        transformed_sample = transform_(sample)
        transform_ = To2DNifti1Image([header_x, header_y])
        transformed_sample = transform_(transformed_sample)

        assert_that(transformed_sample.x, instance_of(nib.Nifti1Image))
        assert_that(transformed_sample.y, instance_of(nib.Nifti1Image))

        NiftiToDisk([self.NIFTI_2D_IMAGE_FROM_SAMPLE, self.NIFTI_2D_LABEL_FROM_SAMPLE]).__call__(transformed_sample)


class RandomCropTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    OUTPUT_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/generated/RandomCrop")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1.nii")
    VALID_MASK_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")
    CROPPED_NIFTI_IMAGE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "cropped_nd_array.nii")
    CROPPED_NIFTI_LABELS = os.path.join(OUTPUT_DATA_FOLDER_PATH, "cropped_labels.nii")
    CROPPED_PNG_IMAGE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "cropped_nd_array.png")
    CROPPED_PNG_LABELS = os.path.join(OUTPUT_DATA_FOLDER_PATH, "cropped_labels.png")

    ALL = [CROPPED_NIFTI_IMAGE, CROPPED_NIFTI_LABELS, CROPPED_PNG_IMAGE, CROPPED_PNG_LABELS]

    @classmethod
    def _flatten(cls, l):
        for el in l:
            if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
                yield from cls._flatten(el)
            else:
                yield el

    def setUp(self):
        pass

    @classmethod
    def setUpClass(cls):
        all_files = cls._flatten(cls.ALL)
        for element in all_files:
            if os.path.exists(element):
                os.remove(element)

        if not os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH):
            os.makedirs(cls.OUTPUT_DATA_FOLDER_PATH)

        assert_that(len(os.listdir(cls.OUTPUT_DATA_FOLDER_PATH)), is_(0))
        assert_that(os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH), is_(True))

    def test_transformer_should_return_ndarray(self):
        nd_array = ToNumpyArray().__call__(self.VALID_3D_NIFTI_FILE)
        labels = ToNumpyArray().__call__(self.VALID_MASK_FILE)

        sample = Sample(x=nd_array, y=labels, is_labeled=True)

        transform_ = transforms.Compose([RandomCrop(output_size=32)])

        transformed_sample = transform_(sample)

        assert_that(transformed_sample.x, instance_of(np.ndarray))
        assert_that(transformed_sample.y, instance_of(np.ndarray))
        assert_that(transformed_sample.x.ndim, equal_to(3))
        assert_that(transformed_sample.y.ndim, equal_to(3))
        assert_that(transformed_sample.x.shape[0], equal_to(1))
        assert_that(transformed_sample.y.shape[0], equal_to(1))

    def test_transformer_should_save_files_as_nifti_for_inspection(self):
        nd_array = ToNumpyArray().__call__(self.VALID_3D_NIFTI_FILE)
        labels = ToNumpyArray().__call__(self.VALID_MASK_FILE)
        sample = Sample(x=nd_array, y=labels, is_labeled=True)
        file_names = [self.CROPPED_NIFTI_IMAGE, self.CROPPED_NIFTI_LABELS]
        transform_ = transforms.Compose([RandomCrop(output_size=32),
                                         To2DNifti1Image(),
                                         NiftiToDisk(
                                             [os.path.join(self.OUTPUT_DATA_FOLDER_PATH, file_path) for file_path in
                                              file_names])])
        transform_(sample)

        for file_name in file_names:
            assert_that(os.path.exists(os.path.join(self.OUTPUT_DATA_FOLDER_PATH, file_name)))

    def test_transformer_should_save_files_as_png_for_inspection(self):
        nd_array = ToNumpyArray().__call__(self.VALID_3D_NIFTI_FILE)
        labels = ToNumpyArray().__call__(self.VALID_MASK_FILE)
        sample = Sample(x=nd_array, y=labels, is_labeled=True)
        file_names = [self.CROPPED_PNG_IMAGE, self.CROPPED_PNG_LABELS]

        transform_ = transforms.Compose([RandomCrop(output_size=32),
                                         ToPNGFile(
                                             [os.path.join(self.OUTPUT_DATA_FOLDER_PATH, file_path) for file_path in
                                              file_names])])
        transform_(sample)

        for file_name in file_names:
            assert_that(os.path.exists(os.path.join(self.OUTPUT_DATA_FOLDER_PATH, file_name)))


class RandomCrop3DTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    OUTPUT_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/generated/RandomCrop3D")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1.nii")
    VALID_MASK_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")
    CROPPED_NIFTI_IMAGE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "cropped_nd_array.nii")
    CROPPED_NIFTI_LABELS = os.path.join(OUTPUT_DATA_FOLDER_PATH, "cropped_labels.nii")
    CROPPED_PNG_IMAGE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "cropped_nd_array.png")
    CROPPED_PNG_LABELS = os.path.join(OUTPUT_DATA_FOLDER_PATH, "cropped_labels.png")

    ALL = [CROPPED_NIFTI_IMAGE, CROPPED_NIFTI_LABELS, CROPPED_PNG_IMAGE, CROPPED_PNG_LABELS]

    @classmethod
    def _flatten(cls, l):
        for el in l:
            if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
                yield from cls._flatten(el)
            else:
                yield el

    def setUp(self):
        pass

    @classmethod
    def setUpClass(cls):
        all_files = cls._flatten(cls.ALL)
        for element in all_files:
            if os.path.exists(element):
                os.remove(element)

        if not os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH):
            os.makedirs(cls.OUTPUT_DATA_FOLDER_PATH)

        assert_that(len(os.listdir(cls.OUTPUT_DATA_FOLDER_PATH)), is_(0))
        assert_that(os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH), is_(True))

    def test_transformer_should_return_ndarray(self):
        nd_array = ToNumpyArray().__call__(self.VALID_3D_NIFTI_FILE)
        labels = ToNumpyArray().__call__(self.VALID_MASK_FILE)
        sample = Sample(x=nd_array, y=labels, is_labeled=True)
        transform_ = transforms.Compose([RandomCrop3D(output_size=32)])

        transformed_sample = transform_(sample)

        assert_that(transformed_sample.x, instance_of(np.ndarray))
        assert_that(transformed_sample.y, instance_of(np.ndarray))
        assert_that(transformed_sample.x.ndim, equal_to(4))
        assert_that(transformed_sample.y.ndim, equal_to(4))
        assert_that(transformed_sample.x.shape[0], equal_to(1))
        assert_that(transformed_sample.y.shape[0], equal_to(1))

    def test_transformer_should_save_files_as_nifti_for_inspection(self):
        nd_array = ToNumpyArray().__call__(self.VALID_3D_NIFTI_FILE)
        labels = ToNumpyArray().__call__(self.VALID_MASK_FILE)
        sample = Sample(x=nd_array, y=labels, is_labeled=True)
        file_names = [self.CROPPED_NIFTI_IMAGE, self.CROPPED_NIFTI_LABELS]

        transform_ = transforms.Compose([RandomCrop3D(output_size=32),
                                         ToNifti1Image([None, None]),
                                         NiftiToDisk(
                                             [os.path.join(self.OUTPUT_DATA_FOLDER_PATH, file_path) for file_path in
                                              file_names])])
        transform_(sample)

        for file_name in file_names:
            assert_that(os.path.exists(os.path.join(self.OUTPUT_DATA_FOLDER_PATH, file_name)))


class ToPNGFileTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    OUTPUT_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/generated/ToPNGFile")
    PNG_FILE_2D = os.path.join(OUTPUT_DATA_FOLDER_PATH, "ToPNGFile2D.png")
    PNG_FILE_3D = os.path.join(OUTPUT_DATA_FOLDER_PATH, "ToPNGFile3D.png")
    PNG_FILE_4D = os.path.join(OUTPUT_DATA_FOLDER_PATH, "ToPNGFile4D.png")
    PNG_SAMPLE_2D = [os.path.join(OUTPUT_DATA_FOLDER_PATH, "ToPNGFileTupleHxW_1.png"),
                     os.path.join(OUTPUT_DATA_FOLDER_PATH, "ToPNGFileTupleHxW_2.png")]
    PNG_SAMPLE_3D = [os.path.join(OUTPUT_DATA_FOLDER_PATH, "ToPNGFileTupleCxHxW_1.png"),
                     os.path.join(OUTPUT_DATA_FOLDER_PATH, "ToPNGFileTupleCxHxW_2.png")]
    PNG_SAMPLE_4D = [os.path.join(OUTPUT_DATA_FOLDER_PATH, "ToPNGFileTupleCxDxHxW_1.png"),
                     os.path.join(OUTPUT_DATA_FOLDER_PATH, "ToPNGFileTupleCxDxHxW_2.png")]
    ALL = [PNG_FILE_2D, PNG_FILE_3D, PNG_FILE_4D, PNG_SAMPLE_2D, PNG_SAMPLE_3D, PNG_SAMPLE_4D]

    @classmethod
    def _flatten(cls, l):
        for el in l:
            if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
                yield from cls._flatten(el)
            else:
                yield el

    def setUp(self):
        pass

    @classmethod
    def setUpClass(cls):
        all_files = cls._flatten(cls.ALL)
        for element in all_files:
            if os.path.exists(element):
                os.remove(element)
        if not os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH):
            os.makedirs(cls.OUTPUT_DATA_FOLDER_PATH)

        assert_that(len(os.listdir(cls.OUTPUT_DATA_FOLDER_PATH)), is_(0))
        assert_that(os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH), is_(True))

    def test_should_raise_exception_when_passing_4D_ndarray(self):
        nd_array = np.random.randint(0, 1000, size=(1, 32, 32, 32))

        transform_ = ToPNGFile(self.PNG_FILE_4D)

        assert_that(calling(transform_.__call__).with_args(nd_array), raises(TypeError))

    def test_should_raise_exception_when_passing_Sample_of_4D_ndarray(self):
        x = np.random.randint(0, 1000, size=(1, 32, 32, 32))
        y = np.random.randint(0, 1000, size=(1, 32, 32, 32))

        sample = Sample(x=x, y=y, is_labeled=True)

        transform_ = ToPNGFile(self.PNG_SAMPLE_4D)

        assert_that(calling(transform_.__call__).with_args(sample), raises(TypeError))

    def test_should_pass_when_passing_2D_ndarray(self):
        nd_array = np.random.randint(0, 1000, size=(32, 32))

        ToPNGFile(self.PNG_FILE_2D).__call__(nd_array)

        assert_that(os.path.exists(os.path.join(self.OUTPUT_DATA_FOLDER_PATH, self.PNG_FILE_2D)))

    def test_should_pass_when_passing_3D_ndarray(self):
        nd_array = np.random.randint(0, 1000, size=(1, 32, 32))

        ToPNGFile(self.PNG_FILE_3D).__call__(nd_array)

        assert_that(os.path.exists(os.path.join(self.OUTPUT_DATA_FOLDER_PATH, self.PNG_FILE_3D)))

    def test_should_pass_when_passing_sample_2D_nd_array(self):

        x = np.random.randint(0, 1000, size=(32, 32))
        y = np.random.randint(0, 1000, size=(32, 32))

        sample = Sample(x=x, y=y, is_labeled=True)

        ToPNGFile(self.PNG_SAMPLE_2D).__call__(sample)

        for file_name in self.PNG_SAMPLE_2D:
            assert_that(os.path.exists(os.path.join(self.OUTPUT_DATA_FOLDER_PATH, file_name)))

    def test_should_pass_when_passing_tuple_3D_nd_array(self):

        x = np.random.randint(0, 1000, size=(1, 32, 32))
        y = np.random.randint(0, 1000, size=(1, 32, 32))

        sample = Sample(x=x, y=y, is_labeled=True)

        ToPNGFile(self.PNG_SAMPLE_3D).__call__(sample)

        for file_name in self.PNG_SAMPLE_3D:
            assert_that(os.path.exists(os.path.join(self.OUTPUT_DATA_FOLDER_PATH, file_name)))


class CropToContentTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    OUTPUT_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/generated/CropToContent")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1.nii")
    VALID_MASK_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")
    CROPPED_NIFTI_IMAGE_FROM_SAMPLE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "cropped_image_sample.nii")
    CROPPED_NIFTI_LABELS_FROM_SAMPLE = os.path.join(OUTPUT_DATA_FOLDER_PATH, "cropped_labels_sample.nii")
    CROPPED_NIFTI_IMAGE_FROM_NUMPY_NDARRAY = os.path.join(OUTPUT_DATA_FOLDER_PATH, "cropped_image_ndarray.nii")
    ALL = [CROPPED_NIFTI_IMAGE_FROM_NUMPY_NDARRAY, CROPPED_NIFTI_IMAGE_FROM_SAMPLE, CROPPED_NIFTI_LABELS_FROM_SAMPLE]

    @classmethod
    def _flatten(cls, l):
        for el in l:
            if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
                yield from cls._flatten(el)
            else:
                yield el

    def setUp(self):
        pass

    @classmethod
    def setUpClass(cls):
        all_files = cls._flatten(cls.ALL)
        for element in all_files:
            if os.path.exists(element):
                os.remove(element)
        if not os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH):
            os.makedirs(cls.OUTPUT_DATA_FOLDER_PATH)

        assert_that(len(os.listdir(cls.OUTPUT_DATA_FOLDER_PATH)), is_(0))
        assert_that(os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH), is_(True))

    def test_should_produce_cropped_to_content_from_sample(self):
        image = nib.load(self.VALID_3D_NIFTI_FILE)
        label = nib.load(self.VALID_MASK_FILE)

        image_data = np.expand_dims(image.get_data().__array__(), axis=-1).transpose((3, 2, 1, 0))
        label_data = np.expand_dims(label.get_data().__array__(), axis=-1).transpose((3, 2, 1, 0))

        sample = Sample(x=image_data, y=label_data, is_labeled=True)

        cropped_sample = CropToContent().__call__(sample)

        assert_that(cropped_sample.x.ndim, equal_to(image_data.ndim))
        assert_that(cropped_sample.x.shape, less_than_or_equal_to(sample.x.shape))

        nifti_image = nib.Nifti1Image(cropped_sample.x.transpose((3, 2, 1, 0)), None, image.header)
        nib.save(nifti_image, self.CROPPED_NIFTI_IMAGE_FROM_SAMPLE)
        nifti_image = nib.Nifti1Image(cropped_sample.y.transpose((3, 2, 1, 0)), None, label.header)
        nib.save(nifti_image, self.CROPPED_NIFTI_LABELS_FROM_SAMPLE)

    def test_should_produce_cropped_to_content_from_ndarray(self):
        image = nib.load(self.VALID_3D_NIFTI_FILE)
        image_data = image.get_data().__array__()
        image_data = np.expand_dims(image_data, axis=-1).transpose((3, 2, 1, 0))
        cropped = CropToContent().__call__(image_data)
        assert_that(cropped.ndim, equal_to(image_data.ndim))
        assert_that(cropped.shape, less_than_or_equal_to(image_data.shape))

        image = nib.Nifti1Image(cropped.transpose((3, 2, 1, 0)), None, image.header)
        nib.save(image, self.CROPPED_NIFTI_IMAGE_FROM_NUMPY_NDARRAY)


class ToTensorPatchesTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    OUTPUT_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/generated/ToTensorPatches")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1_1mm.nii")
    VALID_MASK_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")
    PATCH_SIZE = (1, 96, 128, 128)
    WRONG_PATCH_SIZE = (1, 256, 384, 384)
    STEP = (1, 96, 128, 128)

    def setUp(self) -> None:
        pass

    @classmethod
    def setUpClass(cls):
        if not os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH):
            os.makedirs(cls.OUTPUT_DATA_FOLDER_PATH)

        assert_that(os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH), is_(True))

    def test_should_slice_image(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE)

        transform_ = transforms.Compose([ToNumpyArray(),
                                         ToNDTensor(),
                                         ToTensorPatches(self.PATCH_SIZE, self.STEP)])

        transformed_sample = transform_.__call__(sample)

        assert_that(transformed_sample.x.ndimension(), is_(5))

    def test_should_produce_patches_from_tensor(self):
        transform_ = transforms.Compose([ToNumpyArray()])

        x = torch.from_numpy(transform_.__call__(self.VALID_3D_NIFTI_FILE))

        transformed_x = ToTensorPatches(self.PATCH_SIZE, self.STEP).__call__(x)

        assert_that(transformed_x.ndimension(), is_(5))

    def test_should_raise_value_error_with_patch_too_big(self):
        transform_ = transforms.Compose([ToNumpyArray()])

        x = torch.from_numpy(transform_.__call__(self.VALID_3D_NIFTI_FILE))

        transform_ = ToTensorPatches(self.WRONG_PATCH_SIZE, self.STEP)

        assert_that(calling(transform_).with_args(x), raises(ValueError))

    def test_should_produce_patches_from_sample(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE)

        transform_ = transforms.Compose([ToNumpyArray(),
                                         ToNDTensor(),
                                         ToTensorPatches(self.PATCH_SIZE, self.STEP)])

        transformed_sample = transform_.__call__(sample)

        for i in range(transformed_sample.x.shape[0]):
            img = nib.Nifti1Image(transformed_sample.x[i, 0].numpy(), None)
            nib.save(img, os.path.join(self.OUTPUT_DATA_FOLDER_PATH, "patch_{}".format(i)))


class ToNDArrayPatchesTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    OUTPUT_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/generated/ToNDArrayPatches")
    VALID_3D_NIFTI_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "T1_1mm.nii")
    VALID_MASK_FILE = os.path.join(TEST_DATA_FOLDER_PATH, "Mask.nii")
    PATCH_SIZE = (1, 96, 128, 128)
    WRONG_PATCH_SIZE = (1, 256, 384, 384)
    STEP = (1, 96, 128, 128)

    def setUp(self) -> None:
        pass

    @classmethod
    def setUpClass(cls):
        if not os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH):
            os.makedirs(cls.OUTPUT_DATA_FOLDER_PATH)

        assert_that(os.path.exists(cls.OUTPUT_DATA_FOLDER_PATH), is_(True))

    def test_should_slice_image(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE)

        transform_ = transforms.Compose([ToNumpyArray(),
                                         ToNDArrayPatches(self.PATCH_SIZE, self.STEP)])

        transformed_sample = transform_.__call__(sample)

        assert_that(transformed_sample.x.ndim, is_(5))

    def test_should_produce_patches_from_ndarray(self):
        transform_ = transforms.Compose([ToNumpyArray(),
                                         ToNDArrayPatches(self.PATCH_SIZE, self.STEP)])

        transformed_x = transform_(self.VALID_3D_NIFTI_FILE)

        assert_that(transformed_x.ndim, is_(5))

    def test_should_raise_value_error_with_patch_too_big(self):
        transform_ = transforms.Compose([ToNumpyArray()])

        x = transform_.__call__(self.VALID_3D_NIFTI_FILE)

        transform_ = ToNDArrayPatches(self.WRONG_PATCH_SIZE, self.STEP)

        assert_that(calling(transform_).with_args(x), raises(ValueError))

    def test_should_produce_patches_from_sample(self):
        sample = Sample(x=self.VALID_3D_NIFTI_FILE)

        transform_ = transforms.Compose([ToNumpyArray(),
                                         ToNDArrayPatches(self.PATCH_SIZE, self.STEP)])

        transformed_sample = transform_.__call__(sample)

        for i in range(transformed_sample.x.shape[0]):
            img = nib.Nifti1Image(transformed_sample.x[i, 0], None)
            nib.save(img, os.path.join(self.OUTPUT_DATA_FOLDER_PATH, "patch_{}".format(i)))



#  -*- coding: utf-8 -*-
#  Copyright 2019 SAMITorch Authors. All Rights Reserved.
#  #
#  Licensed under the MIT License;
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://opensource.org/licenses/MIT
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#  ==============================================================================
import torch
from samitorch.inputs.sample import Sample
from typing import List


class Batch(object):
    """
    Represent a Batch during a training process.
    """

    def __init__(self, samples: List[Sample]):
        """
        Batch initializer.

        Args:
            samples: A list of SAMITorch samples.
        """
        self._samples = samples
        self._x = torch.stack([sample.x for sample in samples])
        self._y = torch.stack([sample.y for sample in samples])
        self._dataset_id = torch.cat(
            [torch.tensor([sample.dataset_id], dtype=torch.int8) if sample.dataset_id is not None
             else torch.tensor([0], dtype=torch.int8) for sample in samples])
        self._device = self._x.device

    @property
    def device(self):
        return self._device

    def to_device(self, device):
        self._x = self._x.to(device)
        self._y = self._y.to(device)
        self._dataset_id = self._dataset_id.to(device)
        self._device = device
        return self

    @property
    def samples(self):
        """
        List of samples.

        Returns:
            list: The list of samples.
        """
        return self._samples

    @property
    def x(self):
        """
        Tensor of inputs.

        Returns:
            :obj:`torch.Tensor`: The list of inputs (X).
        """
        return self._x

    @property
    def y(self):
        """
        Tensor of labels.

        Returns:
            :obj:`torch.Tensor`: The list of labels (y).
        """
        return self._y

    @property
    def dataset_id(self):
        """
        Tensor of data set IDs.

        Returns:
            :obj:`torch.Tensor`: The list of data set IDs.
        """
        return self._dataset_id

    @x.setter
    def x(self, x):
        self._x = x

    @y.setter
    def y(self, y):
        self._y = y

    @dataset_id.setter
    def dataset_id(self, dataset_id):
        self._dataset_id = dataset_id

    def update(self, batch, device):
        """
        Update an existing batch from another Batch.

        Args:
            batch (:obj:`samitorch.inputs.batch.Batch`): An updated batch.
            device (:obj:`torch.device`): The target device.

        Returns:
            :obj:`samitorch.inputs.batch.Batch`: The updated Batch.
        """
        self._x = batch.x.to(device)
        self._y = batch.y.to(device)
        self._dataset_id = batch.dataset_id.to(device)
        self._device = device
        return self

    def unpack(self) -> tuple:
        """
        Unpack a Sample.

        Returns:
            tuple: A Tuple of elements representing the (X, y) properties of the Sample.
        """
        return self._x, self._y, self._dataset_id

    @classmethod
    def from_batch(cls, batch):
        """
        Create a new Batch from an existing Batch passed in parameter.

        Args:
            batch (:obj:`samitorch.inputs.batch.Batch`): A template Batch.

        Returns:
            :obj:`samitorch.inputs.batch.Batch`: A new Batch object with same properties as the one passed in
                parameter.
        """
        return cls(batch.samples)



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A base initializer class.

Declare methods an Initializer must have.
"""

import abc


class Initializer(object):

    def __init__(self):
        """Class constructor."""

    pass

    @abc.abstractmethod
    def initialize(self, *args, **kwargs):
        """Initialize a PyTorch torch.nn layer/operation.

        Raises:
            NotImplementedError: If not overwritten by subclass.
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import abc
import datetime

from typing import Union

from samitorch.training.trainer import Trainer
from samitorch.training.model_trainer import ModelTrainer
from samitorch.utils.model_io import save


class TrainingStrategy(object):

    def __init__(self, trainer):
        self._trainer = self._register_trainer(trainer)
        pass

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        pass

    @staticmethod
    def _register_trainer(trainer: Union[ModelTrainer, Trainer]) -> Union[ModelTrainer, Trainer]:
        """Register a trainer to the current training strategy.

        Args:
            trainer (:obj:`Trainer`): A Trainer object.

        Returns:
            Registered trainer.
        """
        assertion_str = "Given trainer is not valid; Must be instance of 'Trainer'."

        assert \
            isinstance(trainer, Trainer), assertion_str

        return trainer


class EarlyStoppingStrategy(TrainingStrategy):
    """Define an early stopping strategy.

    Early stops the training if validation loss doesn't improve after a given patience.
    """

    def __init__(self, patience: int, trainer: Union[ModelTrainer, Trainer]):
        """Class constructor.

        Args:
            patience (int): The number of events (epochs) to count before triggering an action.
            trainer (:obj:`Trainer`): A trainer on which to trigger an action.
        """
        super(EarlyStoppingStrategy).__init__(trainer)
        self._patience = self._register_patience(patience)
        self._counter = 0
        self._best_score = None

    def __call__(self, loss: float):
        """Verify if given metric is better than older one. If so, increase a counter. When counter equals to
        desired patience, finalise the training process.

        Args:
            loss (float): An evaluation score.
        """
        if self._best_score is None:
            self._best_score = loss
        elif loss <= self._best_score:
            self._counter += 1
            if self._counter >= self._patience:
                self._trainer.finalize()
        else:
            self._best_score = loss
            self._counter = 0

    @staticmethod
    def _register_patience(patience: int) -> int:
        """Register a patience to the current training strategy.

        Args:
            patience (int): The number of events to wait before stopping the training process.

        Returns:
            Registered patience.
        """
        assertion_str = "Given patience is not valid; Must be instance of 'int'."
        assertion_str_greater_or_equal = "Given patience is not valid; Must be greater or equal to 1."

        assert isinstance(patience, int), assertion_str
        assert patience >= 1, assertion_str_greater_or_equal

        return patience


class CheckpointStrategy(TrainingStrategy):
    """Define a checkpoint strategy.

    Checkpoint strategy is declared to periodically save a Trainer's model(s).

    """

    def __init__(self, trainer):
        """Class constructor.

        Args:
            trainer (:obj:`Trainer`): A trainer.
        """
        super(CheckpointStrategy).__init__(trainer)

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        """Call method.

        Args:
            *args:
            **kwargs:

        Raises:
            NotImplementedError: if not overwritten by subclass.
        """
        raise NotImplementedError()


class MetricCheckpointStrategy(CheckpointStrategy):
    """Define a checkpoint strategy based on an evaluation score.

    This strategy checks if the metric in parameter is best seen. If so, save the model. If not, simply pass.
    """

    def __init__(self, trainer: Union[ModelTrainer, Trainer], model_name: str):
        super(CheckpointStrategy).__init__(trainer)
        self._best_score = None
        self._model_name = model_name

    def __call__(self, metric: float):
        """Verify if the given metric in parameter is better than an older one. If so, save the model.

        Args:
            metric (float): An evalution score.
        """
        if self._best_score is None:
            self._best_score = metric
        elif metric <= self._best_score:
            pass
        else:
            self._best_score = metric
            time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
            save("{}-{}.pytorch".format(self._model_name, time), self._trainer.config.model, self._trainer.epoch,
                 self._trainer.config.optimizer)


class LossCheckpointStrategy(CheckpointStrategy):
    """Define a checkpoint strategy based on a loss value.

    This strategy checks if the loss in parameter is best seen. If so, save the model. If not, simply pass.
    """

    def __init__(self, trainer: Union[ModelTrainer, Trainer], model_name: str):
        super(CheckpointStrategy).__init__(trainer)
        self._best_score = None
        self._model_name = model_name

    def __call__(self, loss: float):
        """Verify if the given loss in parameter is better than an older one. If so, save the model.

        Args:
            loss (float): A loss value.
        """
        if self._best_score is None:
            self._best_score = loss
        elif loss <= self._best_score:
            self._best_score = loss
            time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
            save("{}-{}.pytorch".format(self._model_name, time), self._trainer.config.model, self._trainer.epoch,
                 self._trainer.config.optimizer)
        else:
            pass



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utility class to declare attributes a Trainer object must have for training a model.
"""

import torch
from torch.utils import data
from typing import Union, List
from samitorch.configs.configurations import RunningConfiguration, Configuration


class TrainerConfiguration(Configuration):

    def __init__(self, checkpoint_every: int, max_epoch: int, criterion: Union[List[torch.nn.Module], torch.nn.Module],
                 metric,
                 model: Union[List[torch.nn.Module], torch.nn.Module],
                 optimizer: Union[List[torch.nn.Module], torch.nn.Module],
                 dataloader: Union[List[torch.utils.data.DataLoader], torch.utils.data.DataLoader],
                 running_config: RunningConfiguration) -> None:
        super(TrainerConfiguration, self).__init__()
        self._checkpoint_every = checkpoint_every
        self._max_epoch = max_epoch
        self._criterion = criterion
        self._metric = metric
        self._model = model
        self._optimizer = optimizer
        self._dataloader = dataloader
        self._running_config = running_config

    @property
    def checkpoint_every(self) -> int:
        """
        Return the frequency of checkpoints.

        Returns:
            int: The frequency (in epoch) at which we should save a model checkpoint.
        """
        return self._checkpoint_every

    @property
    def max_epoch(self) -> int:
        """
        The maximum number of epochs.

        Returns:
            int: The max number of epoch.
        """
        return self._max_epoch

    @property
    def criterion(self) -> torch.nn.Module:
        """
        The problem's criterion (loss) function.

        Returns:
            :obj:`torch.nn.Module`: A PyTorch loss function. Can be a custom made function.
        """
        return self._criterion

    @criterion.setter
    def criterion(self, criterion: torch.nn.Module):
        """
        Set a criterion.

        Args:
            criterion (:obj:`torch.nn.Module`): The modified criterion.
        """
        self._criterion = criterion

    @property
    def metric(self):
        """
        The problem's metric.

        Returns:
            :obj:`torch.nn.Module`: A PyTorch metric. Can be a custom made metric.
        """
        return self._metric

    @metric.setter
    def metric(self, metric):
        """
        Set a model.

        Args:
            metric (:obj:`torch.nn.Module`): A modified metric.
        """
        self._metric = metric

    @property
    def model(self):
        """
        The model.

        Returns:
            :obj:`torch.nn.Module`: A PyTorch model implementing the `forward` method.
        """
        return self._model

    @model.setter
    def model(self, model: torch.nn.Module):
        """
        Set a model.

        Args:
            model (:obj:`torch.nn.Module`): A modified model.
        """
        self._model = model

    @property
    def optimizer(self):
        """
        The problem's optimizer.

        Returns:
            :obj:`torch.nn.Module`: A PyTorch optimizer.
        """
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer: torch.optim):
        """
        Set an optimizer.

        Args:
            optimizer (:obj:`torch.optim`): An optimizer.
        """
        self._optimizer = optimizer

    @property
    def dataloader(self):
        """
        A data loader.

        Returns:
            :obj:`torch.utils.data.DataLoader`: A PyTorch dataloader object.
        """
        return self._dataloader

    @property
    def running_config(self):
        """
        The running configuration, which groups multiple execution environment related variables.

        Returns:
            :obj:`samitorch.configs.configurations.RunningConfiguration`: The running configuration.
        """
        return self._running_config



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import torch
import os

from glob import glob

from typing import List, Tuple


def to_onehot(indices: torch.Tensor, num_classes: int) -> torch.Tensor:
    """
    Convert a tensor of indices of any shape `(N, ...)` to a tensor of one-hot indicators of shape
    `(N, num_classes, ...)`.

    Args:
        indices (:obj:`torch.Tensor`): Tensor of indices.
        num_classes (int): The number of classes of the problem.

    Returns:
        :obj:`torch.Tensor`: The one-hot representation of the input tensor.
    """
    onehot = torch.zeros(indices.shape[0], num_classes, *indices.shape[1:], device=indices.device)
    return onehot.scatter_(1, indices.unsqueeze(1), 1)


def flatten(tensor: torch.Tensor) -> torch.Tensor:
    """
    Flattens a given tensor such that the channel axis is first.
    The shapes are transformed as follows: (N, C, D, H, W) -> (C, N * D * H * W)

    Args:
        tensor (:obj:`torch.Tensor`): Tensor to flatten.

    Returns:
        :obj:`torch.Tensor`: The flattened, 1-D tensor.
    """
    C = tensor.size(1)
    # new axis order
    axis_order = (1, 0) + tuple(range(2, tensor.dim()))
    # Transpose: (N, C, D, H, W) -> (C, N, D, H, W)
    transposed = tensor.permute(axis_order).contiguous()
    # Flatten: (C, N, D, H, W) -> (C, N * D * H * W)
    return transposed.view(C, -1)


def split_filename(filepath: str) -> Tuple[str, str, str]:
    """
    Split a filepath into the directory, base, and extension

    Args:
        filepath (str): The base file path.

    Returns:
        Tuple: The complete file path, base path and file extension.
    """
    path = os.path.dirname(filepath)
    filename = os.path.basename(filepath)
    base, ext = os.path.splitext(filename)
    if ext == '.gz':
        base, ext2 = os.path.splitext(base)
        ext = ext2 + ext
    return path, base, ext


def extract_file_paths(path: str, ext='*.nii*') -> List[str]:
    """
    Grab all `ext` files in a directory and sort them for consistency.

    Args:
        path (str): File path.
        ext (str): File's extension to grab.

    Returns:
        list: A list of string containing every file paths.
    """
    file_paths = sorted(glob(os.path.join(path, ext)))
    return file_paths
